# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

import datetime
import importlib.util
from pathlib import Path

project = "Sublime Music"
copyright = f"{datetime.datetime.today().year}, Sumner Evans"
author = "Sumner Evans"
gitlab_url = "https://gitlab.com/sublime-music/sublime-music/"

# Get the version from the package.
module_name = "sublime_music"
init_file = Path(__file__).parent.parent.joinpath("sublime_music/__init__.py").resolve()
spec = importlib.util.spec_from_file_location(module_name, str(init_file))
sublime_music = importlib.util.module_from_spec(spec)
spec.loader.exec_module(sublime_music)

version = release = f"v{sublime_music.__version__}"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "special-members": "__init__",
}
autosectionlabel_prefix_document = True
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "gtk": ("https://lazka.github.io/pgi-docs/Gtk-3.0/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_logo = "../logo/logo.png"
html_theme_options = {
    "logo_only": True,
}

# Edit on GitLab integration
html_context = {
    "display_gitlab": True,
    "gitlab_user": "sublime-music",
    "gitlab_repo": "sublime-music",
    "gitlab_version": "master",
    "conf_py_path": "/docs/",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for manual page output ---------------------------------------

man_pages = [
    (
        "manpages/sublime-music",
        "sublime-music",
        "native GTK *sonic client",
        "Louis-Philippe Véronneau",
        1,
    )
]



from gi.repository import Gtk, Pango


class SongListColumn(Gtk.TreeViewColumn):
    def __init__(
        self,
        header: str,
        text_idx: int,
        bold: bool = False,
        align: float = 0,
        width: int = None,
    ):
        """Represents a column in a song list."""
        renderer = Gtk.CellRendererText(
            xalign=align,
            weight=Pango.Weight.BOLD if bold else Pango.Weight.NORMAL,
            ellipsize=Pango.EllipsizeMode.END,
        )
        renderer.set_fixed_size(width or -1, 35)

        super().__init__(header, renderer, text=text_idx, sensitive=0)
        self.set_resizable(True)
        self.set_expand(not width)



from pathlib import Path
from typing import Union


def resolve_path(*joinpath_args: Union[str, Path]) -> Path:
    roots = (Path(__file__).parent, Path("/usr/share/sublime-music"))
    for root in roots:
        if (fullpath := root.joinpath(*joinpath_args).resolve()).exists():
            return fullpath

    raise FileNotFoundError(
        f"{Path(*joinpath_args)} could not be found in any of the following "
        "directories: {', '.join(roots)}"
    )



#! /usr/bin/env python

"""
Checks for TODO comments and makes sure they have an associated issue. Formats that are
accepted are:

TODO (#1)
TODO (#1)
TODO (project#1)
TODO (namespace/project#1)
TODO (namespace/namespace/project#1)

Additionally, the TODO can be postfixed with ``:``.
"""

import re
import sys
from pathlib import Path
from typing import Pattern

from termcolor import cprint

todo_re = re.compile(r"\s*#\s*TODO:?\s*")
accounted_for_todo = re.compile(r"\s*#\s*TODO:?\s*\(([\w-]+(/[\w-]+)*)?#\d+\)")


def noqa_re(error_id: str = "") -> Pattern:
    return re.compile(rf"#\s*noqa(:\s*{error_id})?\s*\n$")


def eprint(*strings):
    cprint(" ".join(strings), "red", end="", attrs=["bold"])


def check_file(path: Path) -> bool:
    print(f"Checking {path.absolute()}...")  # noqa: T001
    file = path.open()
    valid = True

    for i, line in enumerate(file, start=1):
        if todo_re.match(line) and not accounted_for_todo.match(line):
            eprint(f"{i}: {line}")
            valid = False

    file.close()
    return valid


valid = True
for path in Path("sublime_music").glob("**/*.py"):
    valid &= check_file(path)

for path in Path("tests").glob("**/*.py"):
    valid &= check_file(path)

"""
Checks that the version in the CHANGELOG is the same as the version in ``__init__.py``.
"""
with open(Path("sublime_music/__init__.py")) as f:
    for line in f:
        if line.startswith("__version__"):
            version = eval(line.split()[-1])
            break

with open(Path("CHANGELOG.rst")) as f:
    assert f.readline().strip() == f"v{version}", "Version mismatch"


sys.exit(0 if valid else 1)



import math
from functools import lru_cache, partial
from random import randint
from typing import Any, cast, Dict, List, Tuple

from fuzzywuzzy import fuzz
from gi.repository import Gdk, Gio, GLib, GObject, Gtk, Pango

from ..adapters import AdapterManager, api_objects as API
from ..config import AppConfiguration
from ..ui import util
from ..ui.common import (
    IconButton,
    LoadError,
    SongListColumn,
    SpinnerImage,
)


class EditPlaylistDialog(Gtk.Dialog):
    def __init__(self, parent: Any, playlist: API.Playlist):
        Gtk.Dialog.__init__(self, transient_for=parent, flags=Gtk.DialogFlags.MODAL)

        # HEADER
        self.header = Gtk.HeaderBar()
        self._set_title(playlist.name)

        cancel_button = Gtk.Button(label="Cancel")
        cancel_button.connect("clicked", lambda _: self.close())
        self.header.pack_start(cancel_button)

        self.edit_button = Gtk.Button(label="Edit")
        self.edit_button.get_style_context().add_class("suggested-action")
        self.edit_button.connect(
            "clicked", lambda *a: self.response(Gtk.ResponseType.APPLY)
        )
        self.header.pack_end(self.edit_button)

        self.set_titlebar(self.header)

        content_area = self.get_content_area()
        content_grid = Gtk.Grid(column_spacing=10, row_spacing=10, margin=10)

        make_label = lambda label_text: Gtk.Label(label_text, halign=Gtk.Align.END)

        content_grid.attach(make_label("Playlist Name"), 0, 0, 1, 1)
        self.name_entry = Gtk.Entry(text=playlist.name, hexpand=True)
        self.name_entry.connect("changed", self._on_name_change)
        content_grid.attach(self.name_entry, 1, 0, 1, 1)

        content_grid.attach(make_label("Comment"), 0, 1, 1, 1)
        self.comment_entry = Gtk.Entry(text=playlist.comment, hexpand=True)
        content_grid.attach(self.comment_entry, 1, 1, 1, 1)

        content_grid.attach(make_label("Public"), 0, 2, 1, 1)
        self.public_switch = Gtk.Switch(active=playlist.public, halign=Gtk.Align.START)
        content_grid.attach(self.public_switch, 1, 2, 1, 1)

        delete_button = Gtk.Button(label="Delete")
        delete_button.connect("clicked", lambda *a: self.response(Gtk.ResponseType.NO))
        content_grid.attach(delete_button, 0, 3, 1, 2)

        content_area.add(content_grid)
        self.show_all()

    def _on_name_change(self, entry: Gtk.Entry):
        text = entry.get_text()
        if len(text) > 0:
            self._set_title(text)
        self.edit_button.set_sensitive(len(text) > 0)

    def _set_title(self, playlist_name: str):
        self.header.props.title = f"Edit {playlist_name}"

    def get_data(self) -> Dict[str, Any]:
        return {
            "name": self.name_entry.get_text(),
            "comment": self.comment_entry.get_text(),
            "public": self.public_switch.get_active(),
        }


class PlaylistsPanel(Gtk.Paned):
    """Defines the playlists panel."""

    __gsignals__ = {
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    def __init__(self, *args, **kwargs):
        Gtk.Paned.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)

        self.playlist_list = PlaylistList()
        self.pack1(self.playlist_list, False, False)

        self.playlist_detail_panel = PlaylistDetailPanel()
        self.playlist_detail_panel.connect(
            "song-clicked",
            lambda _, *args: self.emit("song-clicked", *args),
        )
        self.playlist_detail_panel.connect(
            "refresh-window",
            lambda _, *args: self.emit("refresh-window", *args),
        )
        self.pack2(self.playlist_detail_panel, True, False)

    def update(self, app_config: AppConfiguration = None, force: bool = False):
        self.playlist_list.update(app_config=app_config, force=force)
        self.playlist_detail_panel.update(app_config=app_config, force=force)


class PlaylistList(Gtk.Box):
    __gsignals__ = {
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    offline_mode = False

    class PlaylistModel(GObject.GObject):
        playlist_id = GObject.Property(type=str)
        name = GObject.Property(type=str)

        def __init__(self, playlist_id: str, name: str):
            GObject.GObject.__init__(self)
            self.playlist_id = playlist_id
            self.name = name

    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        playlist_list_actions = Gtk.ActionBar()

        self.new_playlist_button = IconButton("list-add-symbolic", label="New Playlist")
        self.new_playlist_button.connect("clicked", self.on_new_playlist_clicked)
        playlist_list_actions.pack_start(self.new_playlist_button)

        self.list_refresh_button = IconButton(
            "view-refresh-symbolic", "Refresh list of playlists"
        )
        self.list_refresh_button.connect("clicked", self.on_list_refresh_click)
        playlist_list_actions.pack_end(self.list_refresh_button)

        self.add(playlist_list_actions)

        self.error_container = Gtk.Box()
        self.add(self.error_container)

        loading_new_playlist = Gtk.ListBox()

        self.loading_indicator = Gtk.ListBoxRow(activatable=False, selectable=False)
        loading_spinner = Gtk.Spinner(name="playlist-list-spinner", active=True)
        self.loading_indicator.add(loading_spinner)
        loading_new_playlist.add(self.loading_indicator)

        self.new_playlist_row = Gtk.ListBoxRow(activatable=False, selectable=False)
        new_playlist_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=False)

        self.new_playlist_entry = Gtk.Entry(name="playlist-list-new-playlist-entry")
        self.new_playlist_entry.connect("activate", self.new_entry_activate)
        new_playlist_box.add(self.new_playlist_entry)

        new_playlist_actions = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        confirm_button = IconButton(
            "object-select-symbolic",
            "Create playlist",
            name="playlist-list-new-playlist-confirm",
            relief=True,
        )
        confirm_button.connect("clicked", self.confirm_button_clicked)
        new_playlist_actions.pack_end(confirm_button, False, True, 0)

        self.cancel_button = IconButton(
            "process-stop-symbolic",
            "Cancel create playlist",
            name="playlist-list-new-playlist-cancel",
            relief=True,
        )
        self.cancel_button.connect("clicked", self.cancel_button_clicked)
        new_playlist_actions.pack_end(self.cancel_button, False, True, 0)

        new_playlist_box.add(new_playlist_actions)
        self.new_playlist_row.add(new_playlist_box)

        loading_new_playlist.add(self.new_playlist_row)
        self.add(loading_new_playlist)

        list_scroll_window = Gtk.ScrolledWindow(min_content_width=220)

        def create_playlist_row(model: PlaylistList.PlaylistModel) -> Gtk.ListBoxRow:
            row = Gtk.ListBoxRow(
                action_name="app.go-to-playlist",
                action_target=GLib.Variant("s", model.playlist_id),
            )
            row.add(
                Gtk.Label(
                    label=f"<b>{model.name}</b>",
                    use_markup=True,
                    margin=10,
                    halign=Gtk.Align.START,
                    ellipsize=Pango.EllipsizeMode.END,
                )
            )
            row.show_all()
            return row

        self.playlists_store = Gio.ListStore()
        self.list = Gtk.ListBox(name="playlist-list-listbox")
        self.list.bind_model(self.playlists_store, create_playlist_row)
        list_scroll_window.add(self.list)
        self.pack_start(list_scroll_window, True, True, 0)

    def update(self, app_config: AppConfiguration = None, force: bool = False):
        if app_config:
            self.offline_mode = app_config.offline_mode
            self.new_playlist_button.set_sensitive(not app_config.offline_mode)
            self.list_refresh_button.set_sensitive(not app_config.offline_mode)
        self.new_playlist_row.hide()
        self.update_list(app_config=app_config, force=force)

    @util.async_callback(
        AdapterManager.get_playlists,
        before_download=lambda self: self.loading_indicator.show_all(),
        on_failure=lambda self, e: self.loading_indicator.hide(),
    )
    def update_list(
        self,
        playlists: List[API.Playlist],
        app_config: AppConfiguration = None,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        for c in self.error_container.get_children():
            self.error_container.remove(c)
        if is_partial:
            load_error = LoadError(
                "Playlist list",
                "load playlists",
                has_data=len(playlists) > 0,
                offline_mode=self.offline_mode,
            )
            self.error_container.pack_start(load_error, True, True, 0)
            self.error_container.show_all()
        else:
            self.error_container.hide()

        new_store = []
        selected_idx = None
        for i, playlist in enumerate(playlists or []):
            if (
                app_config
                and app_config.state
                and app_config.state.selected_playlist_id == playlist.id
            ):
                selected_idx = i

            new_store.append(PlaylistList.PlaylistModel(playlist.id, playlist.name))

        util.diff_model_store(self.playlists_store, new_store)

        # Preserve selection
        if selected_idx is not None:
            row = self.list.get_row_at_index(selected_idx)
            self.list.select_row(row)

        self.loading_indicator.hide()

    # Event Handlers
    # =========================================================================
    def on_new_playlist_clicked(self, _):
        self.new_playlist_entry.set_text("Untitled Playlist")
        self.new_playlist_entry.grab_focus()
        self.new_playlist_row.show()

    def on_list_refresh_click(self, _):
        self.update(force=True)

    def new_entry_activate(self, entry: Gtk.Entry):
        self.create_playlist(entry.get_text())

    def cancel_button_clicked(self, _):
        self.new_playlist_row.hide()

    def confirm_button_clicked(self, _):
        self.create_playlist(self.new_playlist_entry.get_text())

    def create_playlist(self, playlist_name: str):
        def on_playlist_created(_):
            self.update(force=True)

        self.loading_indicator.show()
        playlist_ceate_future = AdapterManager.create_playlist(name=playlist_name)
        playlist_ceate_future.add_done_callback(
            lambda f: GLib.idle_add(on_playlist_created, f)
        )


class PlaylistDetailPanel(Gtk.Overlay):
    __gsignals__ = {
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    playlist_id = None
    playlist_details_expanded = False
    offline_mode = False

    editing_playlist_song_list: bool = False
    reordering_playlist_song_list: bool = False

    def __init__(self):
        Gtk.Overlay.__init__(self, name="playlist-view-overlay")
        self.playlist_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        playlist_info_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.playlist_artwork = SpinnerImage(
            image_name="playlist-album-artwork",
            spinner_name="playlist-artwork-spinner",
            image_size=200,
        )
        playlist_info_box.add(self.playlist_artwork)

        # Name, comment, number of songs, etc.
        playlist_details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        playlist_details_box.pack_start(Gtk.Box(), True, False, 0)

        self.playlist_indicator = self.make_label(name="playlist-indicator")
        playlist_details_box.add(self.playlist_indicator)

        self.playlist_name = self.make_label(name="playlist-name")
        playlist_details_box.add(self.playlist_name)

        self.playlist_comment = self.make_label(name="playlist-comment")
        playlist_details_box.add(self.playlist_comment)

        self.playlist_stats = self.make_label(name="playlist-stats")
        playlist_details_box.add(self.playlist_stats)

        self.play_shuffle_buttons = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL,
            name="playlist-play-shuffle-buttons",
        )

        self.play_all_button = IconButton(
            "media-playback-start-symbolic",
            label="Play All",
            relief=True,
        )
        self.play_all_button.connect("clicked", self.on_play_all_clicked)
        self.play_shuffle_buttons.pack_start(self.play_all_button, False, False, 0)

        self.shuffle_all_button = IconButton(
            "media-playlist-shuffle-symbolic",
            label="Shuffle All",
            relief=True,
        )
        self.shuffle_all_button.connect("clicked", self.on_shuffle_all_button)
        self.play_shuffle_buttons.pack_start(self.shuffle_all_button, False, False, 5)

        playlist_details_box.add(self.play_shuffle_buttons)

        playlist_info_box.pack_start(playlist_details_box, True, True, 0)

        # Action buttons & expand/collapse button
        action_buttons_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.playlist_action_buttons = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=10
        )

        self.download_all_button = IconButton(
            "folder-download-symbolic", "Download all songs in the playlist"
        )
        self.download_all_button.connect(
            "clicked", self.on_playlist_list_download_all_button_click
        )
        self.playlist_action_buttons.add(self.download_all_button)

        self.playlist_edit_button = IconButton("document-edit-symbolic", "Edit paylist")
        self.playlist_edit_button.connect("clicked", self.on_playlist_edit_button_click)
        self.playlist_action_buttons.add(self.playlist_edit_button)

        self.view_refresh_button = IconButton(
            "view-refresh-symbolic", "Refresh playlist info"
        )
        self.view_refresh_button.connect("clicked", self.on_view_refresh_click)
        self.playlist_action_buttons.add(self.view_refresh_button)

        action_buttons_container.pack_start(
            self.playlist_action_buttons, False, False, 10
        )

        action_buttons_container.pack_start(Gtk.Box(), True, True, 0)

        expand_button_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.expand_collapse_button = IconButton(
            "pan-up-symbolic", "Expand playlist details"
        )
        self.expand_collapse_button.connect("clicked", self.on_expand_collapse_click)
        expand_button_container.pack_end(self.expand_collapse_button, False, False, 0)
        action_buttons_container.add(expand_button_container)

        playlist_info_box.pack_end(action_buttons_container, False, False, 5)

        self.playlist_box.add(playlist_info_box)

        self.error_container = Gtk.Box()
        self.playlist_box.add(self.error_container)

        # Playlist songs list
        self.playlist_song_scroll_window = Gtk.ScrolledWindow()

        self.playlist_song_store = Gtk.ListStore(
            bool,  # clickable
            str,  # cache status
            str,  # title
            str,  # album
            str,  # artist
            str,  # duration
            str,  # song ID
        )

        @lru_cache(maxsize=1024)
        def row_score(key: str, row_items: Tuple[str]) -> int:
            return fuzz.partial_ratio(key, " ".join(row_items).lower())

        def playlist_song_list_search_fn(
            store: Gtk.ListStore,
            col: int,
            key: str,
            treeiter: Gtk.TreeIter,
            data: Any = None,
        ) -> bool:
            threshold = math.ceil(math.ceil(len(key) * 0.8) / len(key) * 100)
            return row_score(key.lower(), tuple(store[treeiter][2:5])) < threshold

        self.playlist_songs = Gtk.TreeView(
            model=self.playlist_song_store,
            reorderable=True,
            margin_top=15,
            enable_search=True,
        )
        self.playlist_songs.set_search_equal_func(playlist_song_list_search_fn)
        selection = self.playlist_songs.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        selection.set_select_function(lambda _, model, path, current: model[path[0]][0])

        # Song status column.
        renderer = Gtk.CellRendererPixbuf()
        renderer.set_fixed_size(30, 35)
        column = Gtk.TreeViewColumn("", renderer, icon_name=1)
        column.set_resizable(True)
        self.playlist_songs.append_column(column)

        self.playlist_songs.append_column(SongListColumn("TITLE", 2, bold=True))
        self.playlist_songs.append_column(SongListColumn("ALBUM", 3))
        self.playlist_songs.append_column(SongListColumn("ARTIST", 4))
        self.playlist_songs.append_column(
            SongListColumn("DURATION", 5, align=1, width=40)
        )

        self.playlist_songs.connect("row-activated", self.on_song_activated)
        self.playlist_songs.connect("button-press-event", self.on_song_button_press)

        # Set up drag-and-drop on the song list for editing the order of the
        # playlist.
        self.playlist_song_store.connect(
            "row-inserted", self.on_playlist_model_row_move
        )
        self.playlist_song_store.connect("row-deleted", self.on_playlist_model_row_move)

        self.playlist_song_scroll_window.add(self.playlist_songs)

        self.playlist_box.pack_start(self.playlist_song_scroll_window, True, True, 0)
        self.add(self.playlist_box)

        playlist_view_spinner = Gtk.Spinner(active=True)
        playlist_view_spinner.start()

        self.playlist_view_loading_box = Gtk.Alignment(
            name="playlist-view-overlay", xalign=0.5, yalign=0.5, xscale=0.1, yscale=0.1
        )
        self.playlist_view_loading_box.add(playlist_view_spinner)
        self.add_overlay(self.playlist_view_loading_box)

    update_playlist_view_order_token = 0

    def update(self, app_config: AppConfiguration, force: bool = False):
        # Deselect everything if switching online to offline.
        if self.offline_mode != app_config.offline_mode:
            self.playlist_songs.get_selection().unselect_all()

        self.offline_mode = app_config.offline_mode
        if app_config.state.selected_playlist_id is None:
            self.playlist_box.hide()
            self.playlist_view_loading_box.hide()
        else:
            self.update_playlist_view_order_token += 1
            self.playlist_box.show()
            self.update_playlist_view(
                app_config.state.selected_playlist_id,
                app_config=app_config,
                force=force,
                order_token=self.update_playlist_view_order_token,
            )
            self.download_all_button.set_sensitive(not app_config.offline_mode)
            self.playlist_edit_button.set_sensitive(not app_config.offline_mode)
            self.view_refresh_button.set_sensitive(not app_config.offline_mode)

    _current_song_ids: List[str] = []

    @util.async_callback(
        AdapterManager.get_playlist_details,
        before_download=lambda self: self.show_loading_all(),
        on_failure=lambda self, e: self.hide_loading_all(),
    )
    def update_playlist_view(
        self,
        playlist: API.Playlist,
        app_config: AppConfiguration = None,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        if self.update_playlist_view_order_token != order_token:
            return

        # If the selected playlist has changed, then clear the selections in
        # the song list.
        if self.playlist_id != playlist.id:
            self.playlist_songs.get_selection().unselect_all()

        self.playlist_id = playlist.id

        if app_config:
            self.playlist_details_expanded = app_config.state.playlist_details_expanded

        up_down = "up" if self.playlist_details_expanded else "down"
        self.expand_collapse_button.set_icon(f"pan-{up_down}-symbolic")
        self.expand_collapse_button.set_tooltip_text(
            "Collapse" if self.playlist_details_expanded else "Expand"
        )

        # Update the info display.
        self.playlist_name.set_markup(f"<b>{playlist.name}</b>")
        self.playlist_name.set_tooltip_text(playlist.name)

        if self.playlist_details_expanded:
            self.playlist_artwork.get_style_context().remove_class("collapsed")
            self.playlist_name.get_style_context().remove_class("collapsed")
            self.playlist_box.show_all()
            self.playlist_indicator.set_markup("PLAYLIST")

            if playlist.comment:
                self.playlist_comment.set_text(playlist.comment)
                self.playlist_comment.set_tooltip_text(playlist.comment)
                self.playlist_comment.show()
            else:
                self.playlist_comment.hide()

            self.playlist_stats.set_markup(self._format_stats(playlist))
        else:
            self.playlist_artwork.get_style_context().add_class("collapsed")
            self.playlist_name.get_style_context().add_class("collapsed")
            self.playlist_box.show_all()
            self.playlist_indicator.hide()
            self.playlist_comment.hide()
            self.playlist_stats.hide()

        # Update the artwork.
        self.update_playlist_artwork(playlist.cover_art, order_token=order_token)

        for c in self.error_container.get_children():
            self.error_container.remove(c)
        if is_partial:
            has_data = len(playlist.songs) > 0
            load_error = LoadError(
                "Playlist data",
                "load playlist details",
                has_data=has_data,
                offline_mode=self.offline_mode,
            )
            self.error_container.pack_start(load_error, True, True, 0)
            self.error_container.show_all()
            if not has_data:
                self.playlist_song_scroll_window.hide()
        else:
            self.error_container.hide()
            self.playlist_song_scroll_window.show()

        # Update the song list model. This requires some fancy diffing to
        # update the list.
        self.editing_playlist_song_list = True

        # This doesn't look efficient, since it's doing a ton of passses over the data,
        # but there is some annoying memory overhead for generating the stores to diff,
        # so we are short-circuiting by checking to see if any of the the IDs have
        # changed.
        #
        # The entire algorithm ends up being O(2n), but the first loop is very tight,
        # and the expensive parts of the second loop are avoided if the IDs haven't
        # changed.
        song_ids, songs = [], []
        if len(self._current_song_ids) != len(playlist.songs):
            force = True

        for i, c in enumerate(playlist.songs):
            if i >= len(self._current_song_ids) or c.id != self._current_song_ids[i]:
                force = True
            song_ids.append(c.id)
            songs.append(c)

        new_songs_store = []
        can_play_any_song = False
        cached_status_icons = ("folder-download-symbolic", "view-pin-symbolic")

        if force:
            self._current_song_ids = song_ids

            # Regenerate the store from the actual song data (this is more expensive
            # because when coming from the cache, we are doing 2N fk requests to
            # albums).
            for status_icon, song in zip(
                util.get_cached_status_icons(song_ids),
                [cast(API.Song, s) for s in songs],
            ):
                playable = not self.offline_mode or status_icon in cached_status_icons
                can_play_any_song |= playable
                new_songs_store.append(
                    [
                        playable,
                        status_icon,
                        song.title,
                        album.name if (album := song.album) else None,
                        artist.name if (artist := song.artist) else None,
                        util.format_song_duration(song.duration),
                        song.id,
                    ]
                )
        else:
            # Just update the clickable state and download state.
            for status_icon, song_model in zip(
                util.get_cached_status_icons(song_ids), self.playlist_song_store
            ):
                playable = not self.offline_mode or status_icon in cached_status_icons
                can_play_any_song |= playable
                new_songs_store.append([playable, status_icon, *song_model[2:]])

        util.diff_song_store(self.playlist_song_store, new_songs_store)

        self.play_all_button.set_sensitive(can_play_any_song)
        self.shuffle_all_button.set_sensitive(can_play_any_song)

        self.editing_playlist_song_list = False

        self.playlist_view_loading_box.hide()
        self.playlist_action_buttons.show_all()

    @util.async_callback(
        partial(AdapterManager.get_cover_art_uri, scheme="file"),
        before_download=lambda self: self.playlist_artwork.set_loading(True),
        on_failure=lambda self, e: self.playlist_artwork.set_loading(False),
    )
    def update_playlist_artwork(
        self,
        cover_art_filename: str,
        app_config: AppConfiguration,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        if self.update_playlist_view_order_token != order_token:
            return

        self.playlist_artwork.set_from_file(cover_art_filename)
        self.playlist_artwork.set_loading(False)

        if self.playlist_details_expanded:
            self.playlist_artwork.set_image_size(200)
        else:
            self.playlist_artwork.set_image_size(70)

    # Event Handlers
    # =========================================================================
    def on_view_refresh_click(self, _):
        self.update_playlist_view(
            self.playlist_id,
            force=True,
            order_token=self.update_playlist_view_order_token,
        )

    def on_playlist_edit_button_click(self, _):
        assert self.playlist_id
        playlist = AdapterManager.get_playlist_details(self.playlist_id).result()
        dialog = EditPlaylistDialog(self.get_toplevel(), playlist)
        playlist_deleted = False

        result = dialog.run()
        # Using ResponseType.NO as the delete event.
        if result not in (Gtk.ResponseType.APPLY, Gtk.ResponseType.NO):
            dialog.destroy()
            return

        if result == Gtk.ResponseType.APPLY:
            AdapterManager.update_playlist(self.playlist_id, **dialog.get_data())
        elif result == Gtk.ResponseType.NO:
            # Delete the playlist.
            confirm_dialog = Gtk.MessageDialog(
                transient_for=self.get_toplevel(),
                message_type=Gtk.MessageType.WARNING,
                buttons=Gtk.ButtonsType.NONE,
                text="Confirm deletion",
            )
            confirm_dialog.add_buttons(
                Gtk.STOCK_DELETE,
                Gtk.ResponseType.YES,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
            )
            confirm_dialog.format_secondary_markup(
                f'Are you sure you want to delete the "{playlist.name}" playlist?'
            )
            result = confirm_dialog.run()
            confirm_dialog.destroy()
            if result == Gtk.ResponseType.YES:
                AdapterManager.delete_playlist(self.playlist_id)
                playlist_deleted = True
            else:
                # In this case, we don't want to do any invalidation of
                # anything.
                dialog.destroy()
                return

        # Force a re-fresh of the view
        self.emit(
            "refresh-window",
            {"selected_playlist_id": None if playlist_deleted else self.playlist_id},
            True,
        )
        dialog.destroy()

    def on_playlist_list_download_all_button_click(self, _):
        def download_state_change(song_id: str):
            GLib.idle_add(
                lambda: self.update_playlist_view(
                    self.playlist_id, order_token=self.update_playlist_view_order_token
                )
            )

        song_ids = [s[-1] for s in self.playlist_song_store]
        AdapterManager.batch_download_songs(
            song_ids,
            before_download=download_state_change,
            on_song_download_complete=download_state_change,
        )

    def on_play_all_clicked(self, _):
        self.emit(
            "song-clicked",
            0,
            [m[-1] for m in self.playlist_song_store],
            {"force_shuffle_state": False, "active_playlist_id": self.playlist_id},
        )

    def on_shuffle_all_button(self, _):
        self.emit(
            "song-clicked",
            randint(0, len(self.playlist_song_store) - 1),
            [m[-1] for m in self.playlist_song_store],
            {"force_shuffle_state": True, "active_playlist_id": self.playlist_id},
        )

    def on_expand_collapse_click(self, _):
        self.emit(
            "refresh-window",
            {"playlist_details_expanded": not self.playlist_details_expanded},
            False,
        )

    def on_song_activated(self, _, idx: Gtk.TreePath, col: Any):
        if not self.playlist_song_store[idx[0]][0]:
            return
        # The song ID is in the last column of the model.
        self.emit(
            "song-clicked",
            idx.get_indices()[0],
            [m[-1] for m in self.playlist_song_store],
            {"active_playlist_id": self.playlist_id},
        )

    def on_song_button_press(self, tree: Gtk.TreeView, event: Gdk.EventButton) -> bool:
        if event.button == 3:  # Right click
            clicked_path = tree.get_path_at_pos(event.x, event.y)
            if not clicked_path:
                return False

            store, paths = tree.get_selection().get_selected_rows()
            allow_deselect = False

            def on_download_state_change(song_id: str):
                GLib.idle_add(
                    lambda: self.update_playlist_view(
                        self.playlist_id,
                        order_token=self.update_playlist_view_order_token,
                    )
                )

            # Use the new selection instead of the old one for calculating what
            # to do the right click on.
            if clicked_path[0] not in paths:
                paths = [clicked_path[0]]
                allow_deselect = True

            song_ids = [self.playlist_song_store[p][-1] for p in paths]

            # Used to adjust for the header row.
            bin_coords = tree.convert_tree_to_bin_window_coords(event.x, event.y)
            widget_coords = tree.convert_tree_to_widget_coords(event.x, event.y)

            def on_remove_songs_click(_):
                assert self.playlist_id
                delete_idxs = {p.get_indices()[0] for p in paths}
                new_song_ids = [
                    model[-1]
                    for i, model in enumerate(self.playlist_song_store)
                    if i not in delete_idxs
                ]
                AdapterManager.update_playlist(
                    playlist_id=self.playlist_id, song_ids=new_song_ids
                ).result()
                self.update_playlist_view(
                    self.playlist_id,
                    force=True,
                    order_token=self.update_playlist_view_order_token,
                )

            remove_text = (
                "Remove " + util.pluralize("song", len(song_ids)) + " from playlist"
            )
            util.show_song_popover(
                song_ids,
                event.x,
                event.y + abs(bin_coords.by - widget_coords.wy),
                tree,
                self.offline_mode,
                on_download_state_change=on_download_state_change,
                on_remove_downloads_click=(
                    lambda: (
                        self.offline_mode
                        and self.playlist_songs.get_selection().unselect_all()
                    )
                ),
                extra_menu_items=[
                    (
                        Gtk.ModelButton(
                            text=remove_text, sensitive=not self.offline_mode
                        ),
                        on_remove_songs_click,
                    )
                ],
                on_playlist_state_change=lambda: self.emit("refresh-window", {}, True),
            )

            # If the click was on a selected row, don't deselect anything.
            if not allow_deselect:
                return True

        return False

    def on_playlist_model_row_move(self, *args):
        # If we are programatically editing the song list, don't do anything.
        if self.editing_playlist_song_list:
            return

        # We get both a delete and insert event, I think it's deterministic
        # which one comes first, but just in case, we have this
        # reordering_playlist_song_list flag.
        if self.reordering_playlist_song_list:
            self._update_playlist_order(self.playlist_id)
            self.reordering_playlist_song_list = False
        else:
            self.reordering_playlist_song_list = True

    # Helper Methods
    # =========================================================================
    def show_loading_all(self):
        self.playlist_artwork.set_loading(True)
        self.playlist_view_loading_box.show_all()

    def hide_loading_all(self):
        self.playlist_artwork.set_loading(False)
        self.playlist_view_loading_box.hide()

    def make_label(self, text: str = None, name: str = None, **params) -> Gtk.Label:
        return Gtk.Label(
            label=text,
            name=name,
            halign=Gtk.Align.START,
            ellipsize=Pango.EllipsizeMode.END,
            **params,
        )

    @util.async_callback(AdapterManager.get_playlist_details)
    def _update_playlist_order(
        self,
        playlist: API.Playlist,
        app_config: AppConfiguration,
        **kwargs,
    ):
        self.playlist_view_loading_box.show_all()
        update_playlist_future = AdapterManager.update_playlist(
            playlist.id, song_ids=[s[-1] for s in self.playlist_song_store]
        )

        update_playlist_future.add_done_callback(
            lambda f: GLib.idle_add(
                lambda: self.update_playlist_view(
                    playlist.id,
                    force=True,
                    order_token=self.update_playlist_view_order_token,
                )
            )
        )

    def _format_stats(self, playlist: API.Playlist) -> str:
        created_date_text = ""
        if playlist.created:
            created_date_text = f" on {playlist.created.strftime('%B %d, %Y')}"
        created_text = f"Created by {playlist.owner}{created_date_text}"

        lines = [
            util.dot_join(
                created_text,
                f"{'Not v' if not playlist.public else 'V'}isible to others",
            ),
            util.dot_join(
                "{} {}".format(
                    playlist.song_count,
                    util.pluralize("song", playlist.song_count or 0),
                ),
                util.format_sequence_duration(playlist.duration),
            ),
        ]
        return "\n".join(lines)



from sublime_music.players.chromecast import ChromecastPlayer


def test_init():
    empty_fn = lambda *a, **k: None
    chromecast_player = ChromecastPlayer(
        empty_fn,
        empty_fn,
        empty_fn,
        empty_fn,
        {
            "Serve Local Files to Chromecasts on the LAN": True,
            "LAN Server Port Number": 6969,
        },
    )
    chromecast_player.shutdown()



import logging
import os
import pickle
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, cast, Dict, Optional, Tuple, Type, Union

import dataclasses_json
from dataclasses_json import config, DataClassJsonMixin

from .adapters import ConfigurationStore
from .ui.state import UIState


# JSON decoder and encoder translations
def encode_path(path: Path) -> str:
    return str(path.resolve())


dataclasses_json.cfg.global_config.decoders[Path] = Path
dataclasses_json.cfg.global_config.decoders[Optional[Path]] = (  # type: ignore
    lambda p: Path(p) if p else None
)


dataclasses_json.cfg.global_config.encoders[Path] = encode_path
dataclasses_json.cfg.global_config.encoders[
    Optional[Path]  # type: ignore
] = encode_path


@dataclass
class ProviderConfiguration:
    id: str
    name: str
    ground_truth_adapter_type: Type
    ground_truth_adapter_config: ConfigurationStore
    caching_adapter_type: Optional[Type] = None
    caching_adapter_config: Optional[ConfigurationStore] = None

    def migrate(self):
        self.ground_truth_adapter_type.migrate_configuration(
            self.ground_truth_adapter_config
        )
        if self.caching_adapter_type:
            self.caching_adapter_type.migrate_configuration(self.caching_adapter_config)

    def clone(self) -> "ProviderConfiguration":
        return ProviderConfiguration(
            self.id,
            self.name,
            self.ground_truth_adapter_type,
            self.ground_truth_adapter_config.clone(),
            self.caching_adapter_type,
            (
                self.caching_adapter_config.clone()
                if self.caching_adapter_config
                else None
            ),
        )

    def persist_secrets(self):
        self.ground_truth_adapter_config.persist_secrets()
        if self.caching_adapter_config:
            self.caching_adapter_config.persist_secrets()


def encode_providers(
    providers_dict: Dict[str, Dict[str, Any]]
) -> Dict[str, Dict[str, Any]]:
    return {
        id_: {
            **config,
            "ground_truth_adapter_type": config["ground_truth_adapter_type"].__name__,
            "caching_adapter_type": (
                cast(type, config.get("caching_adapter_type")).__name__
                if config.get("caching_adapter_type")
                else None
            ),
        }
        for id_, config in providers_dict.items()
    }


def decode_providers(
    providers_dict: Dict[str, Dict[str, Any]]
) -> Dict[str, ProviderConfiguration]:
    from sublime_music.adapters import AdapterManager

    def find_adapter_type(type_name: str) -> Type:
        for available_adapter in AdapterManager.available_adapters:
            if available_adapter.__name__ == type_name:
                return available_adapter
        raise Exception(f"Couldn't find adapter of type {type_name}")

    return {
        id_: ProviderConfiguration(
            config["id"],
            config["name"],
            ground_truth_adapter_type=find_adapter_type(
                config["ground_truth_adapter_type"]
            ),
            ground_truth_adapter_config=ConfigurationStore(
                **config["ground_truth_adapter_config"]
            ),
            caching_adapter_type=(
                find_adapter_type(cat)
                if (cat := config.get("caching_adapter_type"))
                else None
            ),
            caching_adapter_config=(
                ConfigurationStore(**(config.get("caching_adapter_config") or {}))
            ),
        )
        for id_, config in providers_dict.items()
    }


@dataclass
class AppConfiguration(DataClassJsonMixin):
    version: int = 5
    cache_location: Optional[Path] = None
    filename: Optional[Path] = None

    # Providers
    providers: Dict[str, ProviderConfiguration] = field(
        default_factory=dict,
        metadata=config(encoder=encode_providers, decoder=decode_providers),
    )
    current_provider_id: Optional[str] = None
    _loaded_provider_id: Optional[str] = field(default=None, init=False)

    # Players
    player_config: Dict[str, Dict[str, Union[Type, Tuple[str, ...]]]] = field(
        default_factory=dict
    )

    # Global Settings
    song_play_notification: bool = True
    offline_mode: bool = False
    allow_song_downloads: bool = True
    download_on_stream: bool = True  # also download when streaming a song
    prefetch_amount: int = 3
    concurrent_download_limit: int = 5

    # Deprecated. These have also been renamed to avoid using them elsewhere in the app.
    _sol: bool = field(default=True, metadata=config(field_name="serve_over_lan"))
    _pn: int = field(default=8282, metadata=config(field_name="port_number"))
    _rg: int = field(default=0, metadata=config(field_name="replay_gain"))

    @staticmethod
    def load_from_file(filename: Path) -> "AppConfiguration":
        config = AppConfiguration()
        try:
            if filename.exists():
                with open(filename, "r") as f:
                    config = AppConfiguration.from_json(f.read())
        except Exception:
            pass

        config.filename = filename
        return config

    def __post_init__(self):
        # Default the cache_location to ~/.local/share/sublime-music
        if not self.cache_location:
            path = Path(os.environ.get("XDG_DATA_HOME") or "~/.local/share")
            path = path.expanduser().joinpath("sublime-music").resolve()
            self.cache_location = path

        self._state = None
        self._loaded_provider_id = None
        self.migrate()

    def migrate(self):
        for _, provider in self.providers.items():
            provider.migrate()

        if self.version < 6:
            self.player_config = {
                "Local Playback": {"Replay Gain": ["no", "track", "album"][self._rg]},
                "Chromecast": {
                    "Serve Local Files to Chromecasts on the LAN": self._sol,
                    "LAN Server Port Number": self._pn,
                },
            }

        self.version = 6
        self.state.migrate()

    @property
    def provider(self) -> Optional[ProviderConfiguration]:
        return self.providers.get(self.current_provider_id or "")

    @property
    def state(self) -> UIState:
        if not (provider := self.provider):
            return UIState()

        # If the provider has changed, then retrieve the new provider's state.
        if self._loaded_provider_id != provider.id:
            self.load_state()

        return self._state

    def load_state(self):
        self._state = UIState()
        if not (provider := self.provider):
            return

        self._loaded_provider_id = provider.id
        if (state_filename := self._state_file_location) and state_filename.exists():
            try:
                with open(state_filename, "rb") as f:
                    self._state = pickle.load(f)
            except Exception:
                logging.exception(f"Couldn't load state from {state_filename}")
                # Just ignore any errors, it is only UI state.
                self._state = UIState()

        self._state.__init_available_players__()

    @property
    def _state_file_location(self) -> Optional[Path]:
        if not (provider := self.provider):
            return None

        assert self.cache_location
        return self.cache_location.joinpath(provider.id, "state.pickle")

    def save(self):
        # Save the config as YAML.
        self.filename.parent.mkdir(parents=True, exist_ok=True)
        json = self.to_json(indent=2, sort_keys=True)
        with open(self.filename, "w+") as f:
            f.write(json)

        # Save the state for the current provider.
        if state_filename := self._state_file_location:
            state_filename.parent.mkdir(parents=True, exist_ok=True)
            with open(state_filename, "wb+") as f:
                pickle.dump(self.state, f)



from datetime import timedelta
from pathlib import Path

# from time import sleep

from sublime_music.players.mpv import MPVPlayer

MPVPlayer._is_mock = True


def test_init():
    empty_fn = lambda *a, **k: None
    MPVPlayer(empty_fn, empty_fn, empty_fn, empty_fn, {"Replay Gain": "Disabled"})


def is_close(expected: float, value: float, delta: float = 0.5) -> bool:
    print(f"EXPECTED: {expected}, VALUE: {value}")  # noqa: T001
    return abs(value - expected) < delta


def test_play():
    empty_fn = lambda *a, **k: None
    mpv_player = MPVPlayer(
        empty_fn, empty_fn, empty_fn, empty_fn, {"Replay Gain": "Disabled"}
    )

    song_path = Path(__file__).parent.joinpath("mock_data/test-song.mp3")
    mpv_player.play_media(str(song_path), timedelta(seconds=10), None)

    # Test Mute and volume
    # ==================================================================================
    # Test normal volume change.
    assert mpv_player.get_volume() == 100
    mpv_player.set_volume(70)
    assert mpv_player.get_volume() == 70

    # Test mute
    assert not mpv_player.get_is_muted()
    mpv_player.set_muted(True)
    assert mpv_player.get_is_muted()

    # Test volume change when muted
    mpv_player.set_volume(50)
    assert mpv_player.get_volume() == 50
    # The volume of the actual player should still be muted.
    assert mpv_player.mpv.volume == 0

    # Unmute and the volume of the actual player should be what we set (50)
    mpv_player.set_muted(False)
    assert mpv_player.mpv.volume == 50

    # Test Play/Pause
    # ==================================================================================
    # Test Pause
    assert mpv_player.playing
    mpv_player.pause()
    assert not mpv_player.playing

    # Test play
    mpv_player.play()
    assert mpv_player.playing

    # # Test seek
    # for _ in range(10):
    #     sleep(0.1)
    #     if is_close(10, mpv_player.mpv.time_pos):
    #         break
    # else:
    #     raise Exception("Never was close")
    # mpv_player.seek(timedelta(seconds=20))

    # for _ in range(10):
    #     sleep(0.1)
    #     if is_close(20, mpv_player.mpv.time_pos):
    #         break
    # else:
    #     raise Exception("Never was close")

    # Pause so that it doesn't keep playing while testing
    mpv_player.pause()
    mpv_player.shutdown()



import logging
from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from .base import PlayerDeviceEvent, PlayerEvent
from .chromecast import ChromecastPlayer  # noqa: F401
from .mpv import MPVPlayer  # noqa: F401
from ..adapters.api_objects import Song


class PlayerManager:
    # Available Players. Order matters for UI display.
    available_player_types: List[Type] = [MPVPlayer, ChromecastPlayer]

    @staticmethod
    def get_configuration_options() -> Dict[
        str, Dict[str, Union[Type, Tuple[str, ...]]]
    ]:
        """
        :returns: Dictionary of the name of the player -> option configs (see
            :class:`sublime_music.players.base.Player.get_configuration_options` for
            details).
        """
        return {
            p.name: p.get_configuration_options()
            for p in PlayerManager.available_player_types
        }

    # Initialization and Shutdown
    def __init__(
        self,
        on_timepos_change: Callable[[Optional[float]], None],
        on_track_end: Callable[[], None],
        on_player_event: Callable[[PlayerEvent], None],
        player_device_change_callback: Callable[[PlayerDeviceEvent], None],
        config: Dict[str, Dict[str, Union[Type, Tuple[str, ...]]]],
    ):
        self.current_song: Optional[Song] = None
        self.on_timepos_change = on_timepos_change
        self.on_track_end = on_track_end
        self.config = config
        self.players: Dict[Type, Any] = {}
        self.device_id_type_map: Dict[str, Type] = {}
        self._current_device_id: Optional[str] = None

        def player_event_wrapper(pe: PlayerEvent):
            if pe.device_id == self._current_device_id:
                on_player_event(pe)

        self.on_player_event = player_event_wrapper

        def callback_wrapper(pde: PlayerDeviceEvent):
            self.device_id_type_map[pde.id] = pde.player_type
            player_device_change_callback(pde)

        self.player_device_change_callback = callback_wrapper

        self.players = {
            player_type: player_type(
                self.on_timepos_change,
                self.on_track_end,
                self.on_player_event,
                self.player_device_change_callback,
                self.config.get(player_type.name),
            )
            for player_type in PlayerManager.available_player_types
        }

    def change_settings(
        self,
        config: Dict[str, Dict[str, Union[Type, Tuple[str, ...]]]],
    ):
        self.config = config
        for player_type, player in self.players.items():
            player.change_settings(config.get(player_type.name))

    def refresh_players(self):
        for player in self.players.values():
            player.refresh_players()

    def shutdown(self):
        for p in self.players.values():
            p.shutdown()

    def _get_current_player_type(self) -> Any:
        device_id = self._current_device_id
        if device_id:
            return self.device_id_type_map.get(device_id)

    def _get_current_player(self) -> Any:
        if current_player_type := self._get_current_player_type():
            return self.players.get(current_player_type)

    @property
    def supported_schemes(self) -> Set[str]:
        if cp := self._get_current_player():
            return cp.supported_schemes
        return set()

    @property
    def can_start_playing_with_no_latency(self) -> bool:
        if self._current_device_id:
            return self._get_current_player_type().can_start_playing_with_no_latency
        else:
            return False

    @property
    def current_device_id(self) -> Optional[str]:
        return self._current_device_id

    def set_current_device_id(self, device_id: str):
        logging.info(f"Setting current device id to '{device_id}'")
        if cp := self._get_current_player():
            cp.pause()
            cp.song_loaded = False

        self._current_device_id = device_id

        if cp := self._get_current_player():
            cp.set_current_device_id(device_id)
            cp.song_loaded = False

    def reset(self):
        if current_player := self._get_current_player():
            current_player.reset()

    @property
    def song_loaded(self) -> bool:
        if current_player := self._get_current_player():
            return current_player.song_loaded
        return False

    @property
    def playing(self) -> bool:
        if current_player := self._get_current_player():
            return current_player.playing
        return False

    def get_volume(self) -> float:
        if current_player := self._get_current_player():
            return current_player.get_volume()
        return 100

    def set_volume(self, volume: float):
        if current_player := self._get_current_player():
            current_player.set_volume(volume)

    def get_is_muted(self) -> bool:
        if current_player := self._get_current_player():
            return current_player.get_is_muted()
        return False

    def set_muted(self, muted: bool):
        if current_player := self._get_current_player():
            current_player.set_muted(muted)

    def play_media(self, uri: str, progress: timedelta, song: Song):
        self.current_song = song
        if current_player := self._get_current_player():
            current_player.play_media(uri, progress, song)

    def pause(self):
        if current_player := self._get_current_player():
            current_player.pause()

    def toggle_play(self):
        if current_player := self._get_current_player():
            if self.playing:
                current_player.pause()
            else:
                current_player.play()

    def seek(self, position: timedelta):
        if current_player := self._get_current_player():
            current_player.seek(position)



from .adapter_base import (
    Adapter,
    AlbumSearchQuery,
    CacheMissError,
    CachingAdapter,
    ConfigurationStore,
    SongCacheStatus,
    UIInfo,
)
from .configure_server_form import ConfigParamDescriptor, ConfigureServerForm
from .manager import AdapterManager, DownloadProgress, Result, SearchResult

__all__ = (
    "Adapter",
    "AdapterManager",
    "AlbumSearchQuery",
    "CacheMissError",
    "CachingAdapter",
    "ConfigParamDescriptor",
    "ConfigurationStore",
    "ConfigureServerForm",
    "DownloadProgress",
    "Result",
    "SearchResult",
    "SongCacheStatus",
    "UIInfo",
)



from pathlib import Path
from time import sleep

import pytest

from sublime_music.adapters import (
    AdapterManager,
    ConfigurationStore,
    Result,
    SearchResult,
)
from sublime_music.adapters.filesystem import FilesystemAdapter
from sublime_music.adapters.subsonic import api_objects as SubsonicAPI, SubsonicAdapter
from sublime_music.config import AppConfiguration, ProviderConfiguration


@pytest.fixture
def adapter_manager(tmp_path: Path):
    ConfigurationStore.MOCK = True
    subsonic_config_store = ConfigurationStore(
        server_address="https://subsonic.example.com",
        username="test",
        verify_cert=True,
        salt_auth=False,
    )
    subsonic_config_store.set_secret("password", "testpass")

    config = AppConfiguration(
        providers={
            "1": ProviderConfiguration(
                id="1",
                name="foo",
                ground_truth_adapter_type=SubsonicAdapter,
                ground_truth_adapter_config=subsonic_config_store,
                caching_adapter_type=FilesystemAdapter,
                caching_adapter_config=ConfigurationStore(),
            )
        },
        current_provider_id="1",
        cache_location=tmp_path,
    )
    AdapterManager.reset(config, lambda *a: None)
    yield
    AdapterManager.shutdown()


def test_result_immediate():
    result = Result(42)
    assert result.data_is_available
    assert result.result() == 42


def test_result_immediate_callback():
    callback_called = True

    def check_done_callback(f: Result):
        nonlocal callback_called
        assert f.result() == 42
        callback_called = True

    result = Result(42)
    result.add_done_callback(check_done_callback)
    assert callback_called


def test_result_future():
    def resolve_later() -> int:
        sleep(0.1)
        return 42

    result = Result(resolve_later)
    assert not result.data_is_available
    assert result.result() == 42
    assert result.data_is_available


def test_result_future_callback():
    def resolve_later() -> int:
        sleep(0.1)
        return 42

    check_done = False

    def check_done_callback(f: Result):
        nonlocal check_done
        assert result.data_is_available
        assert f.result() == 42
        check_done = True

    result = Result(resolve_later)
    result.add_done_callback(check_done_callback)

    # Should take much less than 1 seconds to complete. If the assertion fails, then the
    # check_done_callback failed.
    t = 0
    while not check_done:
        assert t < 1
        t += 0.1
        sleep(0.1)


def test_default_value():
    def resolve_fail() -> int:
        sleep(0.1)
        raise Exception()

    result = Result(resolve_fail, default_value=42)
    assert not result.data_is_available
    assert result.result() == 42
    assert result.data_is_available


def test_cancel():
    def resolve_later() -> int:
        sleep(0.1)
        return 42

    cancel_called = False

    def on_cancel():
        nonlocal cancel_called
        cancel_called = True

    result = Result(resolve_later, on_cancel=on_cancel)
    result.cancel()
    assert cancel_called
    assert not result.data_is_available
    with pytest.raises(Exception):
        result.result()


def test_get_song_details(adapter_manager: AdapterManager):
    # song = AdapterManager.get_song_details("1")
    # print(song)
    # assert 0
    # TODO (#180)
    pass


def test_search_result_sort():
    search_results1 = SearchResult(query="foo")
    search_results1.add_results(
        "artists",
        [
            # boo != foo so low match rate
            SubsonicAPI.ArtistAndArtistInfo(id=str(i), name=f"boo{i}")
            for i in range(30)
        ],
    )

    search_results2 = SearchResult(query="foo")
    search_results1.add_results(
        "artists",
        [
            # foo == foo, so high match rate
            SubsonicAPI.ArtistAndArtistInfo(id=str(i), name=f"foo{i}")
            for i in range(30)
        ],
    )

    # After unioning, the high match rate ones should be first, and only the top 20
    # should be included.
    search_results1.update(search_results2)
    assert [a.name for a in search_results1.artists] == [f"foo{i}" for i in range(20)]


def test_search_result_update():
    search_results1 = SearchResult(query="foo")
    search_results1.add_results(
        "artists",
        [
            SubsonicAPI.ArtistAndArtistInfo(id="1", name="foo"),
            SubsonicAPI.ArtistAndArtistInfo(id="2", name="another foo"),
        ],
    )

    search_results2 = SearchResult(query="foo")
    search_results2.add_results(
        "artists",
        [SubsonicAPI.ArtistAndArtistInfo(id="3", name="foo2")],
    )

    search_results1.update(search_results2)
    assert [a.name for a in search_results1.artists] == ["foo", "another foo", "foo2"]


def test_search(adapter_manager: AdapterManager):
    # TODO (#180)
    return
    results = []

    # TODO (#180) ingest data

    def search_callback(result: SearchResult):
        results.append((result.artists, result.albums, result.songs, result.playlists))

    AdapterManager.search("ohea", search_callback=search_callback).result()

    # TODO (#180) test getting results from the server and updating using that
    while len(results) < 1:
        sleep(0.1)

    assert len(results) == 1



#! /usr/bin/env python3
import argparse
import logging
import os
from pathlib import Path

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # noqa: F401

import sublime_music

from .app import SublimeMusicApp


def main():
    parser = argparse.ArgumentParser(description="Sublime Music")
    parser.add_argument(
        "-v", "--version", help="show version and exit", action="store_true"
    )
    parser.add_argument("-l", "--logfile", help="the filename to send logs to")
    parser.add_argument(
        "-m",
        "--loglevel",
        help="the minimum level of logging to do",
        default="WARNING",
    )
    parser.add_argument(
        "-c",
        "--config",
        help="specify a configuration file. Defaults to ~/.config/sublime-music/config.json",  # noqa: 512
    )

    args, unknown_args = parser.parse_known_args()
    if args.version:
        print(f"Sublime Music v{sublime_music.__version__}")  # noqa: T001
        return

    min_log_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(min_log_level, int):
        print(f"Invalid log level: {args.loglevel.upper()}.")  # noqa: T001
        min_log_level = logging.WARNING

    logging.basicConfig(
        filename=args.logfile,
        level=min_log_level,
        format="%(asctime)s:%(levelname)s:%(name)s:%(module)s:%(message)s",
    )

    # Config File
    config_file = args.config
    if not config_file:
        # Default to ~/.config/sublime-music.
        config_file = (
            Path(
                os.environ.get("XDG_CONFIG_HOME")
                or os.environ.get("APPDATA")
                or os.path.join("~/.config")
            )
            .joinpath("sublime-music", "config.json")
            .expanduser()
            .resolve()
        )

    app = SublimeMusicApp(Path(config_file))
    app.run(unknown_args)



"""
This file contains all of the classes related for a shared server configuration form.
"""

from dataclasses import dataclass
from functools import partial
from pathlib import Path
from time import sleep
from typing import Any, Callable, cast, Dict, Iterable, Optional, Tuple, Type, Union

from gi.repository import GLib, GObject, Gtk, Pango

from . import ConfigurationStore


@dataclass
class ConfigParamDescriptor:
    """
    Describes a parameter that can be used to configure an adapter. The
    :class:`description`, :class:`required` and :class:`default:` should be self-evident
    as to what they do.

    The :class:`helptext` parameter is optional detailed text that will be shown in a
    help bubble corresponding to the field.

    The :class:`type` must be one of the following:

    * The literal type ``str``: corresponds to a freeform text entry field in the UI.
    * The literal type ``bool``: corresponds to a toggle in the UI.
    * The literal type ``int``: corresponds to a numeric input in the UI.
    * The literal string ``"password"``: corresponds to a password entry field in the
      UI.
    * The literal string ``"option"``: corresponds to dropdown in the UI.
    * The literal type ``Path``: corresponds to a file picker in the UI.

    The :class:`advanced` parameter specifies whether the setting should be behind an
    "Advanced" expander.

    The :class:`numeric_bounds` parameter only has an effect if the :class:`type` is
    `int`. It specifies the min and max values that the UI control can have.

    The :class:`numeric_step` parameter only has an effect if the :class:`type` is
    `int`. It specifies the step that will be taken using the "+" and "-" buttons on the
    UI control (if supported).

    The :class:`options` parameter only has an effect if the :class:`type` is
    ``"option"``. It specifies the list of options that will be available in the
    dropdown in the UI.

    The :class:`pathtype` parameter only has an effect if the :class:`type` is
    ``Path``. It can be either ``"file"`` or ``"directory"`` corresponding to a file
    picker and a directory picker, respectively.
    """

    type: Union[Type, str]
    description: str
    required: bool = True
    helptext: Optional[str] = None
    advanced: Optional[bool] = None
    default: Any = None
    numeric_bounds: Optional[Tuple[int, int]] = None
    numeric_step: Optional[int] = None
    options: Optional[Iterable[str]] = None
    pathtype: Optional[str] = None


class ConfigureServerForm(Gtk.Box):
    __gsignals__ = {
        "config-valid-changed": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (bool,),
        ),
    }

    def __init__(
        self,
        config_store: ConfigurationStore,
        config_parameters: Dict[str, ConfigParamDescriptor],
        verify_configuration: Callable[[], Dict[str, Optional[str]]],
        is_networked: bool = True,
    ):
        """
        Inititialize a :class:`ConfigureServerForm` with the given configuration
        parameters.

        :param config_store: The :class:`ConfigurationStore` to use to store
            configuration values for this adapter.
        :param config_parameters: An dictionary where the keys are the name of the
            configuration paramter and the values are the :class:`ConfigParamDescriptor`
            object corresponding to that configuration parameter. The order of the keys
            in the dictionary correspond to the order that the configuration parameters
            will be shown in the UI.
        :param verify_configuration: A function that verifies whether or not the
            current state of the ``config_store`` is valid. The output should be a
            dictionary containing verification errors. The keys of the returned
            dictionary should be the same as the keys passed in via the
            ``config_parameters`` parameter. The values should be strings describing
            why the corresponding value in the ``config_store`` is invalid.

            If the adapter ``is_networked``, and the special ``"__ping__"`` key is
            returned, then the error will be shown below all of the other settings in
            the ping status box.
        :param is_networked: whether or not the adapter is networked.
        """
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.config_store = config_store
        self.required_config_parameter_keys = set()
        self.verify_configuration = verify_configuration
        self.entries = {}
        self.is_networked = is_networked

        content_grid = Gtk.Grid(
            column_spacing=10,
            row_spacing=5,
            margin_left=10,
            margin_right=10,
        )
        advanced_grid = Gtk.Grid(column_spacing=10, row_spacing=10)

        def create_string_input(is_password: bool, key: str) -> Gtk.Entry:
            entry = Gtk.Entry(
                text=cast(
                    Callable[[str], None],
                    (config_store.get_secret if is_password else config_store.get),
                )(key),
                hexpand=True,
            )
            if is_password:
                entry.set_visibility(False)

            entry.connect(
                "changed",
                lambda e: self._on_config_change(key, e.get_text(), secret=is_password),
            )
            return entry

        def create_bool_input(key: str) -> Gtk.Switch:
            switch = Gtk.Switch(active=config_store.get(key), halign=Gtk.Align.START)
            switch.connect(
                "notify::active",
                lambda s, _: self._on_config_change(key, s.get_active()),
            )
            return switch

        def create_int_input(key: str) -> Gtk.SpinButton:
            raise NotImplementedError()

        def create_option_input(key: str) -> Gtk.ComboBox:
            raise NotImplementedError()

        def create_path_input(key: str) -> Gtk.FileChooser:
            raise NotImplementedError()

        content_grid_i = 0
        advanced_grid_i = 0
        for key, cpd in config_parameters.items():
            if cpd.required:
                self.required_config_parameter_keys.add(key)
            if cpd.default is not None:
                config_store[key] = config_store.get(key, cpd.default)

            label = Gtk.Label(label=cpd.description, halign=Gtk.Align.END)

            input_el_box = Gtk.Box()
            self.entries[key] = cast(
                Callable[[str], Gtk.Widget],
                {
                    str: partial(create_string_input, False),
                    "password": partial(create_string_input, True),
                    bool: create_bool_input,
                    int: create_int_input,
                    "option": create_option_input,
                    Path: create_path_input,
                }[cpd.type],
            )(key)
            input_el_box.add(self.entries[key])

            if cpd.helptext:
                help_icon = Gtk.Image.new_from_icon_name(
                    "help-about",
                    Gtk.IconSize.BUTTON,
                )
                help_icon.get_style_context().add_class("configure-form-help-icon")
                help_icon.set_tooltip_markup(cpd.helptext)
                input_el_box.add(help_icon)

            if not cpd.advanced:
                content_grid.attach(label, 0, content_grid_i, 1, 1)
                content_grid.attach(input_el_box, 1, content_grid_i, 1, 1)
                content_grid_i += 1
            else:
                advanced_grid.attach(label, 0, advanced_grid_i, 1, 1)
                advanced_grid.attach(input_el_box, 1, advanced_grid_i, 1, 1)
                advanced_grid_i += 1

        # Add a button and revealer for the advanced section of the configuration.
        if advanced_grid_i > 0:
            advanced_component = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

            advanced_expander = Gtk.Revealer()
            advanced_expander_icon = Gtk.Image.new_from_icon_name(
                "go-down-symbolic", Gtk.IconSize.BUTTON
            )
            revealed = False

            def toggle_expander(*args):
                nonlocal revealed
                revealed = not revealed
                advanced_expander.set_reveal_child(revealed)
                icon_dir = "up" if revealed else "down"
                advanced_expander_icon.set_from_icon_name(
                    f"go-{icon_dir}-symbolic", Gtk.IconSize.BUTTON
                )

            advanced_expander_button = Gtk.Button(relief=Gtk.ReliefStyle.NONE)
            advanced_expander_button_box = Gtk.Box(
                orientation=Gtk.Orientation.HORIZONTAL, spacing=10
            )

            advanced_label = Gtk.Label(
                label="<b>Advanced Settings</b>", use_markup=True
            )
            advanced_expander_button_box.add(advanced_label)
            advanced_expander_button_box.add(advanced_expander_icon)

            advanced_expander_button.add(advanced_expander_button_box)
            advanced_expander_button.connect("clicked", toggle_expander)
            advanced_component.add(advanced_expander_button)

            advanced_expander.add(advanced_grid)
            advanced_component.add(advanced_expander)

            content_grid.attach(advanced_component, 0, content_grid_i, 2, 1)
            content_grid_i += 1

        content_grid.attach(
            Gtk.Separator(name="config-verification-separator"), 0, content_grid_i, 2, 1
        )
        content_grid_i += 1

        self.config_verification_box = Gtk.Box(spacing=10)
        content_grid.attach(self.config_verification_box, 0, content_grid_i, 2, 1)

        self.pack_start(content_grid, False, False, 10)
        self._verification_status_ratchet = 0
        self._verify_config(self._verification_status_ratchet)

    had_all_required_keys = False
    verifying_in_progress = False

    def _set_verification_status(
        self, verifying: bool, is_valid: bool = False, error_text: str = None
    ):
        from sublime_music.ui import util

        if verifying:
            if not self.verifying_in_progress:
                for c in self.config_verification_box.get_children():
                    self.config_verification_box.remove(c)
                self.config_verification_box.add(
                    Gtk.Spinner(active=True, name="verify-config-spinner")
                )
                self.config_verification_box.add(
                    Gtk.Label(
                        label="<b>Verifying configuration...</b>", use_markup=True
                    )
                )
            self.verifying_in_progress = True
        else:
            self.verifying_in_progress = False
            for c in self.config_verification_box.get_children():
                self.config_verification_box.remove(c)

            def set_icon_and_label(icon_name: str, label_text: str):
                self.config_verification_box.add(
                    Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DND)
                )
                label = Gtk.Label(
                    label=label_text,
                    use_markup=True,
                    ellipsize=Pango.EllipsizeMode.END,
                )
                label.set_tooltip_markup(label_text)
                self.config_verification_box.add(label)

            if is_valid:
                set_icon_and_label(
                    "config-ok-symbolic", "<b>Configuration is valid</b>"
                )
            elif escaped := util.esc(error_text):
                set_icon_and_label("config-error-symbolic", escaped)

        self.config_verification_box.show_all()

    def _on_config_change(self, key: str, value: Any, secret: bool = False):
        if secret:
            self.config_store.set_secret(key, value)
        else:
            self.config_store[key] = value
        self._verification_status_ratchet += 1
        self._verify_config(self._verification_status_ratchet)

    def _verify_config(self, ratchet: int):
        self.emit("config-valid-changed", False)

        from sublime_music.adapters import Result

        if self.required_config_parameter_keys.issubset(set(self.config_store.keys())):
            if self._verification_status_ratchet != ratchet:
                return

            self._set_verification_status(True)

            has_empty = False
            if self.had_all_required_keys:
                for key in self.required_config_parameter_keys:
                    if self.config_store.get(key) == "":
                        self.entries[key].get_style_context().add_class("invalid")
                        self.entries[key].set_tooltip_markup("This field is required")
                        has_empty = True
                    else:
                        self.entries[key].get_style_context().remove_class("invalid")
                        self.entries[key].set_tooltip_markup(None)

            self.had_all_required_keys = True
            if has_empty:
                self._set_verification_status(
                    False,
                    error_text="<b>There are missing fields</b>\n"
                    "Please fill out all required fields.",
                )
                return

            def on_verify_result(verification_errors: Dict[str, Optional[str]]):
                if self._verification_status_ratchet != ratchet:
                    return

                if len(verification_errors) == 0:
                    self.emit("config-valid-changed", True)
                    for entry in self.entries.values():
                        entry.get_style_context().remove_class("invalid")
                    self._set_verification_status(False, is_valid=True)
                    return

                for key, entry in self.entries.items():
                    if error_text := verification_errors.get(key):
                        entry.get_style_context().add_class("invalid")
                        entry.set_tooltip_markup(error_text)
                    else:
                        entry.get_style_context().remove_class("invalid")
                        entry.set_tooltip_markup(None)

                self._set_verification_status(
                    False, error_text=verification_errors.get("__ping__")
                )

            def verify_with_delay() -> Dict[str, Optional[str]]:
                sleep(0.75)
                if self._verification_status_ratchet != ratchet:
                    return {}

                return self.verify_configuration()

            errors_result: Result[Dict[str, Optional[str]]] = Result(verify_with_delay)
            errors_result.add_done_callback(
                lambda f: GLib.idle_add(on_verify_result, f.result())
            )



from datetime import timedelta
from functools import partial
from random import randint
from typing import cast, List, Sequence

from gi.repository import Gio, GLib, GObject, Gtk, Pango

from ..adapters import (
    AdapterManager,
    api_objects as API,
    CacheMissError,
    SongCacheStatus,
)
from ..config import AppConfiguration
from ..ui import util
from ..ui.common import AlbumWithSongs, IconButton, LoadError, SpinnerImage


class ArtistsPanel(Gtk.Paned):
    """Defines the arist panel."""

    __gsignals__ = {
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    def __init__(self, *args, **kwargs):
        Gtk.Paned.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)

        self.artist_list = ArtistList()
        self.pack1(self.artist_list, False, False)

        self.artist_detail_panel = ArtistDetailPanel()
        self.artist_detail_panel.connect(
            "song-clicked",
            lambda _, *args: self.emit("song-clicked", *args),
        )
        self.artist_detail_panel.connect(
            "refresh-window",
            lambda _, *args: self.emit("refresh-window", *args),
        )
        self.pack2(self.artist_detail_panel, True, False)

    def update(self, app_config: AppConfiguration, force: bool = False):
        self.artist_list.update(app_config=app_config)
        self.artist_detail_panel.update(app_config=app_config)


class _ArtistModel(GObject.GObject):
    artist_id = GObject.Property(type=str)
    name = GObject.Property(type=str)
    album_count = GObject.Property(type=int)

    def __init__(self, artist: API.Artist):
        GObject.GObject.__init__(self)
        self.artist_id = artist.id
        self.name = artist.name
        self.album_count = artist.album_count or 0


class ArtistList(Gtk.Box):
    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        list_actions = Gtk.ActionBar()

        self.refresh_button = IconButton(
            "view-refresh-symbolic", "Refresh list of artists"
        )
        self.refresh_button.connect("clicked", lambda *a: self.update(force=True))
        list_actions.pack_end(self.refresh_button)

        self.add(list_actions)

        self.error_container = Gtk.Box()
        self.add(self.error_container)

        self.loading_indicator = Gtk.ListBox()
        spinner_row = Gtk.ListBoxRow(activatable=False, selectable=False)
        spinner = Gtk.Spinner(name="artist-list-spinner", active=True)
        spinner_row.add(spinner)
        self.loading_indicator.add(spinner_row)
        self.pack_start(self.loading_indicator, False, False, 0)

        list_scroll_window = Gtk.ScrolledWindow(min_content_width=250)

        def create_artist_row(model: _ArtistModel) -> Gtk.ListBoxRow:
            label_text = [f"<b>{util.esc(model.name)}</b>"]

            album_count = model.album_count
            if album_count:
                label_text.append(
                    "{} {}".format(album_count, util.pluralize("album", album_count))
                )

            row = Gtk.ListBoxRow(
                action_name="app.go-to-artist",
                action_target=GLib.Variant("s", model.artist_id),
            )
            row.add(
                Gtk.Label(
                    label="\n".join(label_text),
                    use_markup=True,
                    margin=12,
                    halign=Gtk.Align.START,
                    ellipsize=Pango.EllipsizeMode.END,
                )
            )
            row.show_all()
            return row

        self.artists_store = Gio.ListStore()
        self.list = Gtk.ListBox(name="artist-list")
        self.list.bind_model(self.artists_store, create_artist_row)
        list_scroll_window.add(self.list)

        self.pack_start(list_scroll_window, True, True, 0)

    _app_config = None

    @util.async_callback(
        AdapterManager.get_artists,
        before_download=lambda self: self.loading_indicator.show_all(),
        on_failure=lambda self, e: self.loading_indicator.hide(),
    )
    def update(
        self,
        artists: Sequence[API.Artist],
        app_config: AppConfiguration = None,
        is_partial: bool = False,
        **kwargs,
    ):
        if app_config:
            self._app_config = app_config
            self.refresh_button.set_sensitive(not app_config.offline_mode)

        for c in self.error_container.get_children():
            self.error_container.remove(c)
        if is_partial:
            load_error = LoadError(
                "Artist list",
                "load artists",
                has_data=len(artists) > 0,
                offline_mode=(
                    self._app_config.offline_mode if self._app_config else False
                ),
            )
            self.error_container.pack_start(load_error, True, True, 0)
            self.error_container.show_all()
        else:
            self.error_container.hide()

        new_store = []
        selected_idx = None
        for i, artist in enumerate(artists):
            if (
                self._app_config
                and self._app_config.state
                and self._app_config.state.selected_artist_id == artist.id
            ):
                selected_idx = i
            new_store.append(_ArtistModel(artist))

        util.diff_model_store(self.artists_store, new_store)

        # Preserve selection
        if selected_idx is not None:
            row = self.list.get_row_at_index(selected_idx)
            self.list.select_row(row)

        self.loading_indicator.hide()


class ArtistDetailPanel(Gtk.Box):
    """Defines the artists list."""

    __gsignals__ = {
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    update_order_token = 0
    artist_details_expanded = False

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            name="artist-detail-panel",
            orientation=Gtk.Orientation.VERTICAL,
            **kwargs,
        )
        self.albums: Sequence[API.Album] = []
        self.artist_id = None

        # Artist info panel
        self.big_info_panel = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, name="artist-info-panel"
        )

        self.artist_artwork = SpinnerImage(
            loading=False,
            image_name="artist-album-artwork",
            spinner_name="artist-artwork-spinner",
            image_size=300,
        )
        self.big_info_panel.pack_start(self.artist_artwork, False, False, 0)

        # Action buttons, name, comment, number of songs, etc.
        artist_details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        artist_details_box.pack_start(Gtk.Box(), True, False, 0)

        self.artist_indicator = self.make_label(name="artist-indicator")
        artist_details_box.add(self.artist_indicator)

        self.artist_name = self.make_label(
            name="artist-name", ellipsize=Pango.EllipsizeMode.END
        )
        artist_details_box.add(self.artist_name)

        self.artist_bio = self.make_label(
            name="artist-bio", justify=Gtk.Justification.LEFT
        )
        self.artist_bio.set_line_wrap(True)
        artist_details_box.add(self.artist_bio)

        self.similar_artists_scrolledwindow = Gtk.ScrolledWindow()
        similar_artists_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.similar_artists_label = self.make_label(name="similar-artists")
        similar_artists_box.add(self.similar_artists_label)

        self.similar_artists_button_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL
        )
        similar_artists_box.add(self.similar_artists_button_box)
        self.similar_artists_scrolledwindow.add(similar_artists_box)

        artist_details_box.add(self.similar_artists_scrolledwindow)

        self.artist_stats = self.make_label(name="artist-stats")
        artist_details_box.add(self.artist_stats)

        self.play_shuffle_buttons = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL,
            name="playlist-play-shuffle-buttons",
        )

        self.play_button = IconButton(
            "media-playback-start-symbolic", label="Play All", relief=True
        )
        self.play_button.connect("clicked", self.on_play_all_clicked)
        self.play_shuffle_buttons.pack_start(self.play_button, False, False, 0)

        self.shuffle_button = IconButton(
            "media-playlist-shuffle-symbolic", label="Shuffle All", relief=True
        )
        self.shuffle_button.connect("clicked", self.on_shuffle_all_button)
        self.play_shuffle_buttons.pack_start(self.shuffle_button, False, False, 5)
        artist_details_box.add(self.play_shuffle_buttons)

        self.big_info_panel.pack_start(artist_details_box, True, True, 0)

        # Action buttons
        action_buttons_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.artist_action_buttons = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=10
        )

        self.download_all_button = IconButton(
            "folder-download-symbolic", "Download all songs by this artist"
        )
        self.download_all_button.connect("clicked", self.on_download_all_click)
        self.artist_action_buttons.add(self.download_all_button)

        self.refresh_button = IconButton("view-refresh-symbolic", "Refresh artist info")
        self.refresh_button.connect("clicked", self.on_view_refresh_click)
        self.artist_action_buttons.add(self.refresh_button)

        action_buttons_container.pack_start(
            self.artist_action_buttons, False, False, 10
        )

        action_buttons_container.pack_start(Gtk.Box(), True, True, 0)

        expand_button_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.expand_collapse_button = IconButton(
            "pan-up-symbolic", "Expand playlist details"
        )
        self.expand_collapse_button.connect("clicked", self.on_expand_collapse_click)
        expand_button_container.pack_end(self.expand_collapse_button, False, False, 0)
        action_buttons_container.add(expand_button_container)

        self.big_info_panel.pack_start(action_buttons_container, False, False, 5)

        self.pack_start(self.big_info_panel, False, True, 0)

        self.error_container = Gtk.Box()
        self.add(self.error_container)

        self.album_list_scrolledwindow = Gtk.ScrolledWindow()
        self.albums_list = AlbumsListWithSongs()
        self.albums_list.connect(
            "song-clicked",
            lambda _, *args: self.emit("song-clicked", *args),
        )
        self.album_list_scrolledwindow.add(self.albums_list)
        self.pack_start(self.album_list_scrolledwindow, True, True, 0)

    def update(self, app_config: AppConfiguration):
        self.artist_id = app_config.state.selected_artist_id
        self.offline_mode = app_config.offline_mode
        if app_config.state.selected_artist_id is None:
            self.big_info_panel.hide()
            self.album_list_scrolledwindow.hide()
            self.play_shuffle_buttons.hide()
        else:
            self.update_order_token += 1
            self.album_list_scrolledwindow.show()
            self.update_artist_view(
                app_config.state.selected_artist_id,
                app_config=app_config,
                order_token=self.update_order_token,
            )
            self.refresh_button.set_sensitive(not self.offline_mode)
            self.download_all_button.set_sensitive(not self.offline_mode)

    @util.async_callback(
        AdapterManager.get_artist,
        before_download=lambda self: self.set_all_loading(True),
        on_failure=lambda self, e: self.set_all_loading(False),
    )
    def update_artist_view(
        self,
        artist: API.Artist,
        app_config: AppConfiguration,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        if order_token != self.update_order_token:
            return

        self.big_info_panel.show_all()

        if app_config:
            self.artist_details_expanded = app_config.state.artist_details_expanded

        up_down = "up" if self.artist_details_expanded else "down"
        self.expand_collapse_button.set_icon(f"pan-{up_down}-symbolic")
        self.expand_collapse_button.set_tooltip_text(
            "Collapse" if self.artist_details_expanded else "Expand"
        )

        self.artist_name.set_markup(util.esc(f"<b>{artist.name}</b>"))
        self.artist_name.set_tooltip_text(artist.name)

        if self.artist_details_expanded:
            self.artist_artwork.get_style_context().remove_class("collapsed")
            self.artist_name.get_style_context().remove_class("collapsed")
            self.artist_indicator.set_text("ARTIST")
            self.artist_stats.set_markup(self.format_stats(artist))

            if artist.biography:
                self.artist_bio.set_markup(util.esc(artist.biography))
                self.artist_bio.show()
            else:
                self.artist_bio.hide()

            if len(artist.similar_artists or []) > 0:
                self.similar_artists_label.set_markup("<b>Similar Artists:</b> ")
                for c in self.similar_artists_button_box.get_children():
                    self.similar_artists_button_box.remove(c)

                for similar_artist in (artist.similar_artists or [])[:5]:
                    self.similar_artists_button_box.add(
                        Gtk.LinkButton(
                            label=similar_artist.name,
                            name="similar-artist-button",
                            action_name="app.go-to-artist",
                            action_target=GLib.Variant("s", similar_artist.id),
                        )
                    )
                self.similar_artists_scrolledwindow.show_all()
            else:
                self.similar_artists_scrolledwindow.hide()
        else:
            self.artist_artwork.get_style_context().add_class("collapsed")
            self.artist_name.get_style_context().add_class("collapsed")
            self.artist_indicator.hide()
            self.artist_stats.hide()
            self.artist_bio.hide()
            self.similar_artists_scrolledwindow.hide()

        self.play_shuffle_buttons.show_all()

        self.update_artist_artwork(
            artist.artist_image_url,
            force=force,
            order_token=order_token,
        )

        for c in self.error_container.get_children():
            self.error_container.remove(c)
        if is_partial:
            has_data = len(artist.albums or []) > 0
            load_error = LoadError(
                "Artist data",
                "load artist details",
                has_data=has_data,
                offline_mode=self.offline_mode,
            )
            self.error_container.pack_start(load_error, True, True, 0)
            self.error_container.show_all()
            if not has_data:
                self.album_list_scrolledwindow.hide()
        else:
            self.error_container.hide()
            self.album_list_scrolledwindow.show()

        self.albums = artist.albums or []

        # (Dis|En)able the "Play All" and "Shuffle All" buttons. If in offline mode, it
        # depends on whether or not there are any cached songs.
        if self.offline_mode:
            has_cached_song = False
            playable_statuses = (
                SongCacheStatus.CACHED,
                SongCacheStatus.PERMANENTLY_CACHED,
            )

            for album in self.albums:
                if album.id:
                    try:
                        songs = AdapterManager.get_album(album.id).result().songs or []
                    except CacheMissError as e:
                        if e.partial_data:
                            songs = cast(API.Album, e.partial_data).songs or []
                        else:
                            songs = []
                    statuses = AdapterManager.get_cached_statuses([s.id for s in songs])
                    if any(s in playable_statuses for s in statuses):
                        has_cached_song = True
                        break

            self.play_button.set_sensitive(has_cached_song)
            self.shuffle_button.set_sensitive(has_cached_song)
        else:
            self.play_button.set_sensitive(not self.offline_mode)
            self.shuffle_button.set_sensitive(not self.offline_mode)

        self.albums_list.update(artist, app_config, force=force)

    @util.async_callback(
        partial(AdapterManager.get_cover_art_uri, scheme="file"),
        before_download=lambda self: self.artist_artwork.set_loading(True),
        on_failure=lambda self, e: self.artist_artwork.set_loading(False),
    )
    def update_artist_artwork(
        self,
        cover_art_filename: str,
        app_config: AppConfiguration,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        if order_token != self.update_order_token:
            return
        self.artist_artwork.set_from_file(cover_art_filename)
        self.artist_artwork.set_loading(False)

        if self.artist_details_expanded:
            self.artist_artwork.set_image_size(300)
        else:
            self.artist_artwork.set_image_size(70)

    # Event Handlers
    # =========================================================================
    def on_view_refresh_click(self, *args):
        self.update_artist_view(
            self.artist_id,
            force=True,
            order_token=self.update_order_token,
        )

    def on_download_all_click(self, _):
        AdapterManager.batch_download_songs(
            self.get_artist_song_ids(),
            before_download=lambda _: self.update_artist_view(
                self.artist_id,
                order_token=self.update_order_token,
            ),
            on_song_download_complete=lambda _: self.update_artist_view(
                self.artist_id,
                order_token=self.update_order_token,
            ),
        )

    def on_play_all_clicked(self, _):
        songs = self.get_artist_song_ids()
        self.emit(
            "song-clicked",
            0,
            songs,
            {"force_shuffle_state": False},
        )

    def on_shuffle_all_button(self, _):
        songs = self.get_artist_song_ids()
        self.emit(
            "song-clicked",
            randint(0, len(songs) - 1),
            songs,
            {"force_shuffle_state": True},
        )

    def on_expand_collapse_click(self, _):
        self.emit(
            "refresh-window",
            {"artist_details_expanded": not self.artist_details_expanded},
            False,
        )

    # Helper Methods
    # =========================================================================
    def set_all_loading(self, loading_state: bool):
        if loading_state:
            self.albums_list.spinner.start()
            self.albums_list.spinner.show()
            self.artist_artwork.set_loading(True)
        else:
            self.albums_list.spinner.hide()
            self.artist_artwork.set_loading(False)

    def make_label(self, text: str = None, name: str = None, **params) -> Gtk.Label:
        return Gtk.Label(
            label=text, name=name, halign=Gtk.Align.START, xalign=0, **params
        )

    def format_stats(self, artist: API.Artist) -> str:
        album_count = artist.album_count or len(artist.albums or [])
        song_count, duration = 0, timedelta(0)
        for album in artist.albums or []:
            song_count += album.song_count or 0
            duration += album.duration or timedelta(0)

        return util.dot_join(
            "{} {}".format(album_count, util.pluralize("album", album_count)),
            "{} {}".format(song_count, util.pluralize("song", song_count)),
            util.format_sequence_duration(duration),
        )

    def get_artist_song_ids(self) -> List[str]:
        try:
            artist = AdapterManager.get_artist(self.artist_id).result()
        except CacheMissError as c:
            artist = cast(API.Artist, c.partial_data)

        if not artist:
            return []

        songs = []
        for album in artist.albums or []:
            assert album.id
            try:
                album_with_songs = AdapterManager.get_album(album.id).result()
            except CacheMissError as c:
                album_with_songs = cast(API.Album, c.partial_data)
            if not album_with_songs:
                continue
            for song in album_with_songs.songs or []:
                songs.append(song.id)

        return songs


class AlbumsListWithSongs(Gtk.Overlay):
    __gsignals__ = {
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
    }

    def __init__(self):
        Gtk.Overlay.__init__(self)
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.add(self.box)

        self.spinner = Gtk.Spinner(
            name="albumslist-with-songs-spinner",
            active=False,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

        self.albums = []

    def update(
        self, artist: API.Artist, app_config: AppConfiguration, force: bool = False
    ):
        def remove_all():
            for c in self.box.get_children():
                self.box.remove(c)

        if artist is None:
            remove_all()
            self.spinner.hide()
            return

        new_albums = sorted(
            artist.albums or [], key=lambda a: (a.year or float("inf"), a.name)
        )

        if self.albums == new_albums:
            # Just go through all of the colidren and update them.
            for c in self.box.get_children():
                c.update(app_config=app_config, force=force)

            self.spinner.hide()
            return

        self.albums = new_albums

        remove_all()

        for album in self.albums:
            album_with_songs = AlbumWithSongs(album, show_artist_name=False)
            album_with_songs.connect(
                "song-clicked",
                lambda _, *args: self.emit("song-clicked", *args),
            )
            album_with_songs.connect("song-selected", self.on_song_selected)
            album_with_songs.show_all()
            self.box.add(album_with_songs)

        # Update everything (no force to ensure that if we are online, then everything
        # is clickable)
        for c in self.box.get_children():
            c.update(app_config=app_config)

        self.spinner.hide()

    def on_song_selected(self, album_component: AlbumWithSongs):
        for child in self.box.get_children():
            if album_component != child:
                child.deselect_all()



from pathlib import Path

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # noqa: F401

from sublime_music.ui import common


def test_icon_buttons():
    common.IconButton("cloud-offline")
    common.IconToggleButton("cloud-offline")
    common.IconMenuButton("cloud-offline")


def test_load_error():
    test_cases = [
        (
            (True, True),
            "cloud-offline",
            "Song list may be incomplete.\nGo online to load song list.",
        ),
        ((True, False), "network-error", "Error attempting to load song list."),
        ((False, True), "cloud-offline", "Go online to load song list."),
        ((False, False), "network-error", "Error attempting to load song list."),
    ]
    for (has_data, offline_mode), icon_name, label_text in test_cases:

        load_error = common.LoadError(
            "Song list", "load song list", has_data=has_data, offline_mode=offline_mode
        )
        assert load_error.image.get_icon_name().icon_name == f"{icon_name}-symbolic"
        assert load_error.label.get_text() == label_text


def test_song_list_column():
    common.SongListColumn("H", 1, bold=True, align=1.0, width=30)


def test_spinner_image():
    initial_size = 300
    image = common.SpinnerImage(
        loading=False,
        image_name="test",
        spinner_name="ohea",
        image_size=initial_size,
    )
    image.set_from_file(None)
    assert image.image.get_pixbuf() is None

    image.set_from_file("")
    assert image.image.get_pixbuf() is None

    image.set_from_file(
        str(Path(__file__).parent.joinpath("mock_data", "album-art.png"))
    )
    assert (pixbuf := image.image.get_pixbuf()) is not None
    assert pixbuf.get_width() == pixbuf.get_height() == initial_size

    smaller_size = 70
    image.set_image_size(smaller_size)
    assert (pixbuf := image.image.get_pixbuf()) is not None
    assert pixbuf.get_width() == pixbuf.get_height() == smaller_size

    # Just make sure these don't raise exceptions.
    image.set_loading(True)
    image.set_loading(False)



from typing import Any, Optional

from gi.repository import Gtk


class IconButton(Gtk.Button):
    def __init__(
        self,
        icon_name: Optional[str],
        tooltip_text: str = "",
        relief: bool = False,
        icon_size: Gtk.IconSize = Gtk.IconSize.BUTTON,
        label: str = None,
        **kwargs,
    ):
        Gtk.Button.__init__(self, **kwargs)

        self.icon_size = icon_size
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, name="icon-button-box")

        self.image = Gtk.Image.new_from_icon_name(icon_name, self.icon_size)
        box.pack_start(self.image, False, False, 0)

        if label is not None:
            box.add(Gtk.Label(label=label))

        if not relief:
            self.props.relief = Gtk.ReliefStyle.NONE

        self.add(box)
        self.set_tooltip_text(tooltip_text)

    def set_icon(self, icon_name: Optional[str]):
        self.image.set_from_icon_name(icon_name, self.icon_size)


class IconToggleButton(Gtk.ToggleButton):
    def __init__(
        self,
        icon_name: Optional[str],
        tooltip_text: str = "",
        relief: bool = False,
        icon_size: Gtk.IconSize = Gtk.IconSize.BUTTON,
        label: str = None,
        **kwargs,
    ):
        Gtk.ToggleButton.__init__(self, **kwargs)
        self.icon_size = icon_size
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, name="icon-button-box")

        self.image = Gtk.Image.new_from_icon_name(icon_name, self.icon_size)
        box.add(self.image)

        if label is not None:
            box.add(Gtk.Label(label=label))

        if not relief:
            self.props.relief = Gtk.ReliefStyle.NONE

        self.add(box)
        self.set_tooltip_text(tooltip_text)

    def set_icon(self, icon_name: Optional[str]):
        self.image.set_from_icon_name(icon_name, self.icon_size)

    def get_active(self) -> bool:
        return super().get_active()

    def set_active(self, active: bool):
        super().set_active(active)


class IconMenuButton(Gtk.MenuButton):
    def __init__(
        self,
        icon_name: Optional[str] = None,
        tooltip_text: str = "",
        relief: bool = True,
        icon_size: Gtk.IconSize = Gtk.IconSize.BUTTON,
        label: str = None,
        popover: Any = None,
        **kwargs,
    ):
        Gtk.MenuButton.__init__(self, **kwargs)

        if popover:
            self.set_use_popover(True)
            self.set_popover(popover)

        self.icon_size = icon_size
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, name="icon-button-box")

        self.image = Gtk.Image.new_from_icon_name(icon_name, self.icon_size)
        box.add(self.image)

        if label is not None:
            box.add(Gtk.Label(label=label))

        self.props.relief = Gtk.ReliefStyle.NORMAL

        self.add(box)
        self.set_tooltip_text(tooltip_text)

    def set_icon(self, icon_name: Optional[str]):
        self.image.set_from_icon_name(icon_name, self.icon_size)

    def set_from_file(self, icon_file: Optional[str]):
        self.image.set_from_file(icon_file)



import hashlib
import json
import logging
import re
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Generator, List, Tuple

import pytest
from dateutil.tz import tzutc

from sublime_music.adapters import ConfigurationStore
from sublime_music.adapters.subsonic import api_objects as SubsonicAPI, SubsonicAdapter

MOCK_DATA_FILES = Path(__file__).parent.joinpath("mock_data")


@pytest.fixture
def adapter(tmp_path: Path):
    ConfigurationStore.MOCK = True
    config = ConfigurationStore(
        server_address="https://subsonic.example.com",
        username="test",
        verify_cert=True,
        salt_auth=False,
    )
    config.set_secret("password", "testpass")

    adapter = SubsonicAdapter(config, tmp_path)
    adapter._is_mock = True

    yield adapter
    adapter.shutdown()


@pytest.fixture
def salt_auth_adapter(tmp_path: Path):
    ConfigurationStore.MOCK = True
    config = ConfigurationStore(
        server_address="https://subsonic.example.com",
        username="test",
        verify_cert=True,
        salt_auth=True,
    )
    config.set_secret("password", "testpass")

    adapter = SubsonicAdapter(config, tmp_path)
    adapter._is_mock = True

    yield adapter
    adapter.shutdown()


def mock_data_files(
    request_name: str, mode: str = "r"
) -> Generator[Tuple[Path, Any], None, None]:
    """
    Yields all of the files, and each of the elements of in the file (separated by a
    line of ='s), for all files in the mock_data directory that start with
    ``request_name``. This only works for text such as JSON.
    """
    sep_re = re.compile(r"=+\n")

    num_files = 0
    for file in MOCK_DATA_FILES.iterdir():
        if file.name.split("-")[0] == request_name:
            with open(file, mode) as f:
                parts: List[str] = []
                aggregate: List[str] = []
                for line in f:
                    if sep_re.match(line):
                        parts.append("\n".join(aggregate))
                        aggregate = []
                        continue
                    aggregate.append(line)

                parts.append("\n".join(aggregate))
                print(file)  # noqa: T001
                num_files += 1
                yield file, iter(parts)

    # Make sure that is at least one test file
    assert num_files > 0


def mock_json(**obj: Any) -> str:
    return json.dumps(
        {"subsonic-response": {"status": "ok", "version": "1.15.0", **obj}}
    )


def camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def test_config_form():
    # Just make sure that the functions work. That's half of the battle.
    config_store = ConfigurationStore()
    SubsonicAdapter.get_configuration_form(config_store)


def test_plain_auth_logic(adapter: SubsonicAdapter):
    expected = {
        "u": "test",
        "p": "testpass",
        "c": "Sublime Music",
        "f": "json",
        "v": "1.8.0",
    }
    assert sorted(expected.items()) == sorted(adapter._get_params().items())


def test_salt_auth_logic(salt_auth_adapter: SubsonicAdapter):
    expected = {"u": "test", "c": "Sublime Music", "f": "json", "v": "1.8.0"}

    params = salt_auth_adapter._get_params()
    assert "p" not in params
    assert "s" in params
    salt = params["s"]
    assert "t" in params
    assert params["t"] == hashlib.md5(f"testpass{salt}".encode()).hexdigest()
    assert all(key in params and params[key] == expected[key] for key in expected)


def test_migrate_configuration_populate_salt_auth():
    config = ConfigurationStore(
        server_address="https://subsonic.example.com",
        username="test",
        verify_cert=True,
    )
    SubsonicAdapter.migrate_configuration(config)
    assert "salt_auth" in config
    assert config["salt_auth"]


def test_migrate_configuration_salt_auth_present():
    config = ConfigurationStore(
        server_address="https://subsonic.example.com",
        username="test",
        verify_cert=True,
        salt_auth=False,
    )
    SubsonicAdapter.migrate_configuration(config)
    assert "salt_auth" in config
    assert not config["salt_auth"]


def test_make_url(adapter: SubsonicAdapter):
    assert adapter._make_url("foo") == "https://subsonic.example.com/rest/foo.view"


def test_ping_status(adapter: SubsonicAdapter):
    # Mock a connection error
    adapter._set_mock_data(Exception())
    assert not adapter.ping_status

    # Simulate some sort of ping error
    for filename, data in mock_data_files("ping_failed"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)
        assert not adapter.ping_status

    # Simulate valid ping
    adapter._set_mock_data(mock_json())
    adapter._last_ping_timestamp.value = 0.0
    adapter._set_ping_status()
    assert adapter.ping_status


def test_get_playlists(adapter: SubsonicAdapter):
    expected = [
        SubsonicAPI.Playlist(
            id="2",
            name="Test",
            song_count=132,
            duration=timedelta(seconds=33072),
            created=datetime(2020, 3, 27, 5, 38, 45, 0, tzinfo=tzutc()),
            changed=datetime(2020, 4, 9, 16, 3, 26, 0, tzinfo=tzutc()),
            comment="Foo",
            owner="foo",
            public=True,
            cover_art="pl-2",
        ),
        SubsonicAPI.Playlist(
            id="3",
            name="Bar",
            song_count=23,
            duration=timedelta(seconds=847),
            created=datetime(2020, 3, 27, 5, 39, 4, 0, tzinfo=tzutc()),
            changed=datetime(2020, 3, 27, 5, 45, 23, 0, tzinfo=tzutc()),
            comment="",
            owner="foo",
            public=False,
            cover_art="pl-3",
        ),
    ]

    for filename, data in mock_data_files("get_playlists"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)
        assert adapter.get_playlists() == sorted(expected, key=lambda e: e.name)

    # When playlists is null, expect an empty list.
    adapter._set_mock_data(mock_json())
    assert adapter.get_playlists() == []


def test_get_playlist_details(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_playlist_details"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        playlist_details = adapter.get_playlist_details("2")

        # Make sure that the song count is correct even if it's not provided.
        # Old versions of Subsonic don't have these properties.
        assert len(playlist_details.songs) == 2
        assert playlist_details.duration == timedelta(seconds=470)

        # Make sure that at least the first song got decoded properly.
        assert playlist_details.songs[0] == SubsonicAPI.Song(
            id="202",
            parent_id="318",
            title="What a Beautiful Name",
            _album="What a Beautiful Name - Single",
            album_id="48",
            _artist="Hillsong Worship",
            artist_id="38",
            track=1,
            year=2016,
            _genre="Christian & Gospel",
            cover_art="318",
            size=8381640,
            duration=timedelta(seconds=238),
            path="/".join(
                (
                    "Hillsong Worship",
                    "What a Beautiful Name - Single",
                    "01 What a Beautiful Name.m4a",
                )
            ),
            disc_number=1,
        )


def test_create_playlist(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("create_playlist"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        adapter.create_playlist(
            name="Foo",
            songs=[
                SubsonicAPI.Song(
                    id="202",
                    parent_id="318",
                    title="What a Beautiful Name",
                    _album="What a Beautiful Name - Single",
                    album_id="48",
                    _artist="Hillsong Worship",
                    artist_id="38",
                    track=1,
                    year=2016,
                    _genre="Christian & Gospel",
                    cover_art="318",
                    duration=timedelta(seconds=238),
                    path="/".join(
                        (
                            "Hillsong Worship",
                            "What a Beautiful Name - Single",
                            "01 What a Beautiful Name.m4a",
                        )
                    ),
                    disc_number=1,
                )
            ],
        )


def test_update_playlist(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("update_playlist"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        result_playlist = adapter.update_playlist(
            "1", name="Foo", comment="Bar", public=True, song_ids=["202"]
        )

        assert result_playlist.comment == "Bar"
        assert result_playlist.public is False


def test_get_song_details(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_song_details"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        song = adapter.get_song_details("1")
        assert (song.id, song.title, song.year, song.cover_art, song.duration) == (
            "1",
            "Sweet Caroline",
            2017,
            "544",
            timedelta(seconds=203),
        )
        assert song.path and song.path.endswith("Sweet Caroline.mp3")
        assert song.parent_id == "544"
        assert song.artist
        assert (song.artist.id, song.artist.name) == ("60", "Neil Diamond")
        assert song.album
        assert (song.album.id, song.album.name) == ("88", "50th Anniversary Collection")
        assert song.genre and song.genre.name == "Pop"


def test_get_song_details_missing_data(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_song_details_no_albumid"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        song = adapter.get_song_details("1")
        assert (song.id, song.title, song.year, song.cover_art, song.duration) == (
            "1",
            "Sweet Caroline",
            2017,
            "544",
            timedelta(seconds=203),
        )
        assert song.path and song.path.endswith("Sweet Caroline.mp3")
        assert song.parent_id == "544"
        assert song.artist
        assert (song.artist.id, song.artist.name) == ("60", "Neil Diamond")
        assert song.album
        assert (song.album.id, song.album.name) == (None, "50th Anniversary Collection")
        assert song.genre and song.genre.name == "Pop"

    for filename, data in mock_data_files("get_song_details_no_artistid"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        song = adapter.get_song_details("1")
        assert (song.id, song.title, song.year, song.cover_art, song.duration) == (
            "1",
            "Sweet Caroline",
            2017,
            "544",
            timedelta(seconds=203),
        )
        assert song.path and song.path.endswith("Sweet Caroline.mp3")
        assert song.parent_id == "544"
        assert song.artist
        assert (song.artist.id, song.artist.name) == (None, "Neil Diamond")
        assert song.album
        assert (song.album.id, song.album.name) == ("88", "50th Anniversary Collection")
        assert song.genre and song.genre.name == "Pop"


def test_get_genres(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_genres"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        genres = adapter.get_genres()

        assert len(genres) == 2
        assert [g.name for g in genres] == ["Country", "Pop"]


def test_get_artists(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_artists"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        artists = adapter.get_artists()
        assert len(artists) == 7
        assert {a.name for a in artists} == {
            "Adele",
            "Austin  French",
            "The Afters",
            "The Band Perry",
            "Basshunter",
            "Zac Brown Band",
            "Zach Williams",
        }


def test_get_ignored_articles(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_artists"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        ignored_articles = adapter.get_ignored_articles()
        assert ignored_articles == {"The", "El", "La", "Los", "Las", "Le", "Les"}


def test_get_ignored_articles_from_cached_get_artists(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_artists"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        adapter.get_artists()
        ignored_articles = adapter.get_ignored_articles()
        assert ignored_articles == {"The", "El", "La", "Los", "Las", "Le", "Les"}


def test_get_artist(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_artist"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        artist = adapter.get_artist("3")
        assert artist.album_count == 1
        assert artist.albums and len(artist.albums) == 1
        assert ("3", "Kane Brown") == (artist.albums[0].id, artist.albums[0].name)
        assert artist.artist_image_url == "ar-3"
        assert artist.biography and len(artist.biography) > 0
        assert artist.name == "Kane Brown"
        assert artist.similar_artists
        assert len(artist.similar_artists) == 20
        assert (first_similar := artist.similar_artists[0])
        assert first_similar
        assert first_similar.name == "Luke Combs"
        assert first_similar.artist_image_url == "ar-158"


def test_get_artist_with_good_image_url(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_artist_good_image_url"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        artist = adapter.get_artist("3")
        assert artist.album_count == 1
        assert artist.albums and len(artist.albums) == 1
        assert artist.biography and len(artist.biography) > 0
        assert artist.name == "Kane Brown"
        assert ("3", "Kane Brown") == (artist.albums[0].id, artist.albums[0].name)
        assert (
            artist.artist_image_url
            == "http://entertainermag.com/wp-content/uploads/2017/04/Kane-Brown-Web-Optimized.jpg"  # noqa: E501
        )


def test_get_play_queue(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_play_queue"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        play_queue = adapter.get_play_queue()
        assert play_queue
        assert play_queue.current_index and play_queue.current_index == 1
        assert play_queue.position == timedelta(milliseconds=98914)
        assert play_queue.username == "sumner"
        assert play_queue.changed == datetime(
            2020, 5, 12, 5, 16, 32, 114000, tzinfo=timezone.utc
        )
        assert play_queue.songs and len(play_queue.songs) == 5

        song = play_queue.songs[0]
        assert song.album and song.album.name == "Despacito"
        assert song.artist and song.artist.name == "Peter Bence"
        assert song.genre and song.genre.name == "Classical"


def test_get_album(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_album"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        album = adapter.get_album("243")
        assert (
            album.id,
            album.name,
            album.cover_art,
            album.song_count,
            album.year,
            album.duration,
        ) == (
            "243",
            "What You See Is What You Get",
            "al-243",
            17,
            2019,
            timedelta(seconds=3576),
        )
        assert album.artist
        assert (album.artist.id, album.artist.name) == ("158", "Luke Combs")
        assert album.genre and album.genre.name == "Country"
        assert album.songs
        assert len(album.songs) == 17
        assert [s.title for s in album.songs] == [
            "Beer Never Broke My Heart",
            "Refrigerator Door",
            "Even Though I'm Leaving",
            "Lovin' On You",
            "Moon Over Mexico",
            "1, 2 Many",
            "Blue Collar Boys",
            "New Every Day",
            "Reasons",
            "Every Little Bit Helps",
            "Dear Today",
            "What You See Is What You Get",
            "Does To Me",
            "Angels Workin' Overtime",
            "All Over Again",
            "Nothing Like You",
            "Better Together",
        ]


def test_get_music_directory(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("get_music_directory"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        directory = adapter.get_directory("3")
        assert directory.id == "60"
        assert directory.name == "Luke Bryan"
        assert directory.parent_id == "root"
        assert directory.children and len(directory.children) == 1
        child = directory.children[0]
        assert isinstance(child, SubsonicAPI.Directory)
        assert child.id == "542"
        assert child.name == "Crash My Party"
        assert child.parent_id == "60"

    for filename, data in mock_data_files("get_indexes"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        directory = adapter.get_directory("root")
        assert directory.id == "root"
        assert directory.parent_id is None
        assert len(directory.children) == 7
        child = directory.children[0]
        assert isinstance(child, SubsonicAPI.Directory)
        assert child.id == "73"
        assert child.name == "The Afters"
        assert child.parent_id == "root"


def test_search(adapter: SubsonicAdapter):
    for filename, data in mock_data_files("search3"):
        logging.info(filename)
        logging.debug(data)
        adapter._set_mock_data(data)

        search_results = adapter.search("3")
        assert len(search_results._songs) == 7
        assert len(search_results._artists) == 2
        assert len(search_results._albums) == 4



import copy
import math
from datetime import timedelta
from functools import partial
from typing import Any, Callable, Dict, Optional, Set, Tuple

from gi.repository import Gdk, GdkPixbuf, GLib, GObject, Gtk, Pango

from . import util
from .common import IconButton, IconToggleButton, SpinnerImage
from .state import RepeatType
from ..adapters import AdapterManager, Result, SongCacheStatus
from ..adapters.api_objects import Song
from ..config import AppConfiguration
from ..util import resolve_path


class PlayerControls(Gtk.ActionBar):
    """
    Defines the player controls panel that appears at the bottom of the window.
    """

    __gsignals__ = {
        "song-scrub": (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (float,)),
        "volume-change": (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (float,)),
        "device-update": (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (str,)),
        "song-clicked": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        "songs-removed": (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (object,)),
        "refresh-window": (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }
    editing: bool = False
    editing_play_queue_song_list: bool = False
    reordering_play_queue_song_list: bool = False
    current_song = None
    current_device = None
    current_playing_index: Optional[int] = None
    current_play_queue: Tuple[str, ...] = ()
    cover_art_update_order_token = 0
    play_queue_update_order_token = 0
    offline_mode = False

    def __init__(self):
        Gtk.ActionBar.__init__(self)
        self.set_name("player-controls-bar")

        song_display = self.create_song_display()
        playback_controls = self.create_playback_controls()
        play_queue_volume = self.create_play_queue_volume()

        self.last_device_list_update = None

        self.pack_start(song_display)
        self.set_center_widget(playback_controls)
        self.pack_end(play_queue_volume)

    connecting_to_device_token = 0
    connecting_icon_index = 0

    def update(self, app_config: AppConfiguration, force: bool = False):
        self.current_device = app_config.state.current_device
        self.update_device_list(app_config)

        duration = (
            app_config.state.current_song.duration
            if app_config.state.current_song
            else None
        )
        song_stream_cache_progress = (
            app_config.state.song_stream_cache_progress
            if app_config.state.current_song
            else None
        )
        self.update_scrubber(
            app_config.state.song_progress, duration, song_stream_cache_progress
        )

        icon = "pause" if app_config.state.playing else "start"
        self.play_button.set_icon(f"media-playback-{icon}-symbolic")
        self.play_button.set_tooltip_text(
            "Pause" if app_config.state.playing else "Play"
        )

        has_current_song = app_config.state.current_song is not None
        has_next_song = False
        if app_config.state.repeat_type in (
            RepeatType.REPEAT_QUEUE,
            RepeatType.REPEAT_SONG,
        ):
            has_next_song = True
        elif has_current_song:
            last_idx_in_queue = len(app_config.state.play_queue) - 1
            has_next_song = app_config.state.current_song_index < last_idx_in_queue

        # Toggle button states.
        self.repeat_button.set_action_name(None)
        self.shuffle_button.set_action_name(None)
        repeat_on = app_config.state.repeat_type in (
            RepeatType.REPEAT_QUEUE,
            RepeatType.REPEAT_SONG,
        )
        self.repeat_button.set_active(repeat_on)
        self.repeat_button.set_icon(app_config.state.repeat_type.icon)
        self.shuffle_button.set_active(app_config.state.shuffle_on)
        self.repeat_button.set_action_name("app.repeat-press")
        self.shuffle_button.set_action_name("app.shuffle-press")

        self.song_scrubber.set_sensitive(has_current_song)
        self.prev_button.set_sensitive(has_current_song)
        self.play_button.set_sensitive(has_current_song)
        self.next_button.set_sensitive(has_current_song and has_next_song)

        self.connecting_to_device = app_config.state.connecting_to_device

        def cycle_connecting(connecting_to_device_token: int):
            if (
                self.connecting_to_device_token != connecting_to_device_token
                or not self.connecting_to_device
            ):
                return
            icon = f"chromecast-connecting-{self.connecting_icon_index}-symbolic"
            self.device_button.set_icon(icon)
            self.connecting_icon_index = (self.connecting_icon_index + 1) % 3
            GLib.timeout_add(350, cycle_connecting, connecting_to_device_token)

        icon = ""
        if app_config.state.connecting_to_device:
            icon = "-connecting-0"
            self.connecting_icon_index = 0
            self.connecting_to_device_token += 1
            GLib.timeout_add(350, cycle_connecting, self.connecting_to_device_token)
        elif app_config.state.current_device != "this device":
            icon = "-connected"

        self.device_button.set_icon(f"chromecast{icon}-symbolic")

        # Volume button and slider
        if app_config.state.is_muted:
            icon_name = "muted"
        elif app_config.state.volume < 30:
            icon_name = "low"
        elif app_config.state.volume < 70:
            icon_name = "medium"
        else:
            icon_name = "high"

        self.volume_mute_toggle.set_icon(f"audio-volume-{icon_name}-symbolic")

        self.editing = True
        self.volume_slider.set_value(
            0 if app_config.state.is_muted else app_config.state.volume
        )
        self.editing = False

        # Update the current song information.
        # TODO (#126): add popup of bigger cover art photo here
        if app_config.state.current_song is not None:
            self.cover_art_update_order_token += 1
            self.update_cover_art(
                app_config.state.current_song.cover_art,
                order_token=self.cover_art_update_order_token,
            )

            self.song_title.set_markup(util.esc(app_config.state.current_song.title))
            # TODO (#71): use walrus once MYPY gets its act together
            album = app_config.state.current_song.album
            artist = app_config.state.current_song.artist
            if album:
                self.album_name.set_markup(util.esc(album.name))
                self.artist_name.show()
            else:
                self.album_name.set_markup("")
                self.album_name.hide()
            if artist:
                self.artist_name.set_markup(util.esc(artist.name))
                self.artist_name.show()
            else:
                self.artist_name.set_markup("")
                self.artist_name.hide()
        else:
            # Clear out the cover art and song tite if no song
            self.album_art.set_from_file(None)
            self.album_art.set_loading(False)
            self.song_title.set_markup("")
            self.album_name.set_markup("")
            self.artist_name.set_markup("")

        self.load_play_queue_button.set_sensitive(not self.offline_mode)
        if app_config.state.loading_play_queue:
            self.play_queue_spinner.start()
            self.play_queue_spinner.show()
        else:
            self.play_queue_spinner.stop()
            self.play_queue_spinner.hide()

        # Short circuit if no changes to the play queue
        force |= self.offline_mode != app_config.offline_mode
        self.offline_mode = app_config.offline_mode
        if not force and (
            self.current_play_queue == app_config.state.play_queue
            and self.current_playing_index == app_config.state.current_song_index
        ):
            return
        self.current_play_queue = app_config.state.play_queue
        self.current_playing_index = app_config.state.current_song_index

        # Set the Play Queue button popup.
        play_queue_len = len(app_config.state.play_queue)
        if play_queue_len == 0:
            self.popover_label.set_markup("<b>Play Queue</b>")
        else:
            song_label = util.pluralize("song", play_queue_len)
            self.popover_label.set_markup(
                f"<b>Play Queue:</b> {play_queue_len} {song_label}"
            )

        # TODO (#207) this is super freaking stupid inefficient.
        # IDEAS: batch it, don't get the queue until requested
        self.editing_play_queue_song_list = True

        new_store = []

        def calculate_label(song_details: Song) -> str:
            title = util.esc(song_details.title)
            # TODO (#71): use walrus once MYPY works with this
            # album = util.esc(album.name if (album := song_details.album) else None)
            # artist = util.esc(artist.name if (artist := song_details.artist) else None)  # noqa
            album = util.esc(song_details.album.name if song_details.album else None)
            artist = util.esc(song_details.artist.name if song_details.artist else None)
            return f"<b>{title}</b>\n{util.dot_join(album, artist)}"

        def make_idle_index_capturing_function(
            idx: int,
            order_tok: int,
            fn: Callable[[int, int, Any], None],
        ) -> Callable[[Result], None]:
            return lambda f: GLib.idle_add(fn, idx, order_tok, f.result())

        def on_cover_art_future_done(
            idx: int,
            order_token: int,
            cover_art_filename: str,
        ):
            if order_token != self.play_queue_update_order_token:
                return

            self.play_queue_store[idx][1] = cover_art_filename

        def get_cover_art_filename_or_create_future(
            cover_art_id: Optional[str], idx: int, order_token: int
        ) -> Optional[str]:
            cover_art_result = AdapterManager.get_cover_art_uri(cover_art_id, "file")
            if not cover_art_result.data_is_available:
                cover_art_result.add_done_callback(
                    make_idle_index_capturing_function(
                        idx, order_token, on_cover_art_future_done
                    )
                )
                return None

            # The cover art is already cached.
            return cover_art_result.result()

        def on_song_details_future_done(idx: int, order_token: int, song_details: Song):
            if order_token != self.play_queue_update_order_token:
                return

            self.play_queue_store[idx][2] = calculate_label(song_details)

            # Cover Art
            filename = get_cover_art_filename_or_create_future(
                song_details.cover_art, idx, order_token
            )
            if filename:
                self.play_queue_store[idx][1] = filename

        current_play_queue = [x[-1] for x in self.play_queue_store]
        if app_config.state.play_queue != current_play_queue:
            self.play_queue_update_order_token += 1

        song_details_results = []
        for i, (song_id, cached_status) in enumerate(
            zip(
                app_config.state.play_queue,
                AdapterManager.get_cached_statuses(app_config.state.play_queue),
            )
        ):
            song_details_result = AdapterManager.get_song_details(song_id)

            cover_art_filename = ""
            label = "\n"

            if song_details_result.data_is_available:
                # We have the details of the song already cached.
                song_details = song_details_result.result()
                label = calculate_label(song_details)

                filename = get_cover_art_filename_or_create_future(
                    song_details.cover_art, i, self.play_queue_update_order_token
                )
                if filename:
                    cover_art_filename = filename
            else:
                song_details_results.append((i, song_details_result))

            new_store.append(
                [
                    (
                        not self.offline_mode
                        or cached_status
                        in (SongCacheStatus.CACHED, SongCacheStatus.PERMANENTLY_CACHED)
                    ),
                    cover_art_filename,
                    label,
                    i == app_config.state.current_song_index,
                    song_id,
                ]
            )

        util.diff_song_store(self.play_queue_store, new_store)

        # Do this after the diff to avoid race conditions.
        for idx, song_details_result in song_details_results:
            song_details_result.add_done_callback(
                make_idle_index_capturing_function(
                    idx,
                    self.play_queue_update_order_token,
                    on_song_details_future_done,
                )
            )

        self.editing_play_queue_song_list = False

    @util.async_callback(
        partial(AdapterManager.get_cover_art_uri, scheme="file"),
        before_download=lambda self: self.album_art.set_loading(True),
        on_failure=lambda self, e: self.album_art.set_loading(False),
    )
    def update_cover_art(
        self,
        cover_art_filename: str,
        app_config: AppConfiguration,
        force: bool = False,
        order_token: int = None,
        is_partial: bool = False,
    ):
        if order_token != self.cover_art_update_order_token:
            return

        self.album_art.set_from_file(cover_art_filename)
        self.album_art.set_loading(False)

    def update_scrubber(
        self,
        current: Optional[timedelta],
        duration: Optional[timedelta],
        song_stream_cache_progress: Optional[timedelta],
    ):
        if current is None or duration is None:
            self.song_duration_label.set_text("-:--")
            self.song_progress_label.set_text("-:--")
            self.song_scrubber.set_value(0)
            return

        percent_complete = current / duration * 100

        if not self.editing:
            self.song_scrubber.set_value(percent_complete)

        self.song_scrubber.set_show_fill_level(song_stream_cache_progress is not None)
        if song_stream_cache_progress is not None:
            percent_cached = song_stream_cache_progress / duration * 100
            self.song_scrubber.set_fill_level(percent_cached)

        self.song_duration_label.set_text(util.format_song_duration(duration))
        self.song_progress_label.set_text(
            util.format_song_duration(math.floor(current.total_seconds()))
        )

    def on_volume_change(self, scale: Gtk.Scale):
        if not self.editing:
            self.emit("volume-change", scale.get_value())

    def on_play_queue_click(self, _: Any):
        if self.play_queue_popover.is_visible():
            self.play_queue_popover.popdown()
        else:
            # TODO (#88): scroll the currently playing song into view.
            self.play_queue_popover.popup()
            self.play_queue_popover.show_all()

            # Hide the load play queue button if the adapter can't do that.
            if not AdapterManager.can_get_play_queue():
                self.load_play_queue_button.hide()

    def on_song_activated(self, t: Any, idx: Gtk.TreePath, c: Any):
        if not self.play_queue_store[idx[0]][0]:
            return
        # The song ID is in the last column of the model.
        self.emit(
            "song-clicked",
            idx.get_indices()[0],
            [m[-1] for m in self.play_queue_store],
            {"no_reshuffle": True},
        )

    _current_player_id = None
    _current_available_players: Dict[type, Set[Tuple[str, str]]] = {}

    def update_device_list(self, app_config: AppConfiguration):
        if (
            self._current_available_players == app_config.state.available_players
            and self._current_player_id == app_config.state.current_device
        ):
            return

        self._current_player_id = app_config.state.current_device
        self._current_available_players = copy.deepcopy(
            app_config.state.available_players
        )
        for c in self.device_list.get_children():
            self.device_list.remove(c)

        for i, (player_type, players) in enumerate(
            app_config.state.available_players.items()
        ):
            if len(players) == 0:
                continue
            if i > 0:
                self.device_list.add(
                    Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
                )
            self.device_list.add(
                Gtk.Label(
                    label=f"{player_type.name} Devices",
                    halign=Gtk.Align.START,
                    name="device-type-section-title",
                )
            )

            for player_id, player_name in sorted(players, key=lambda p: p[1]):
                icon = (
                    "audio-volume-high-symbolic"
                    if player_id == self.current_device
                    else None
                )
                button = IconButton(icon, label=player_name)
                button.get_style_context().add_class("menu-button")
                button.connect(
                    "clicked",
                    lambda _, player_id: self.emit("device-update", player_id),
                    player_id,
                )
                self.device_list.add(button)

        self.device_list.show_all()

    def on_device_click(self, _: Any):
        if self.device_popover.is_visible():
            self.device_popover.popdown()
        else:
            self.device_popover.popup()
            self.device_popover.show_all()

    def on_play_queue_button_press(self, tree: Any, event: Gdk.EventButton) -> bool:
        if event.button == 3:  # Right click
            clicked_path = tree.get_path_at_pos(event.x, event.y)

            store, paths = tree.get_selection().get_selected_rows()
            allow_deselect = False

            def on_download_state_change(song_id: str):
                # Refresh the entire window (no force) because the song could
                # be in a list anywhere in the window.
                self.emit("refresh-window", {}, False)

            # Use the new selection instead of the old one for calculating what
            # to do the right click on.
            if clicked_path[0] not in paths:
                paths = [clicked_path[0]]
                allow_deselect = True

            song_ids = [self.play_queue_store[p][-1] for p in paths]

            remove_text = (
                "Remove " + util.pluralize("song", len(song_ids)) + " from queue"
            )

            def on_remove_songs_click(_: Any):
                self.emit("songs-removed", [p.get_indices()[0] for p in paths])

            util.show_song_popover(
                song_ids,
                event.x,
                event.y,
                tree,
                self.offline_mode,
                on_download_state_change=on_download_state_change,
                extra_menu_items=[
                    (Gtk.ModelButton(text=remove_text), on_remove_songs_click),
                ],
            )

            # If the click was on a selected row, don't deselect anything.
            if not allow_deselect:
                return True

        return False

    def on_play_queue_model_row_move(self, *args):
        # If we are programatically editing the song list, don't do anything.
        if self.editing_play_queue_song_list:
            return

        # We get both a delete and insert event, I think it's deterministic
        # which one comes first, but just in case, we have this
        # reordering_play_queue_song_list flag.
        if self.reordering_play_queue_song_list:
            currently_playing_index = [
                i for i, s in enumerate(self.play_queue_store) if s[3]  # playing
            ][0]
            self.emit(
                "refresh-window",
                {
                    "current_song_index": currently_playing_index,
                    "play_queue": tuple(s[-1] for s in self.play_queue_store),
                },
                False,
            )
            self.reordering_play_queue_song_list = False
        else:
            self.reordering_play_queue_song_list = True

    def create_song_display(self) -> Gtk.Box:
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.album_art = SpinnerImage(
            image_name="player-controls-album-artwork",
            image_size=70,
        )
        box.pack_start(self.album_art, False, False, 0)

        details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        details_box.pack_start(Gtk.Box(), True, True, 0)

        def make_label(name: str) -> Gtk.Label:
            return Gtk.Label(
                name=name,
                halign=Gtk.Align.START,
                xalign=0,
                use_markup=True,
                ellipsize=Pango.EllipsizeMode.END,
            )

        self.song_title = make_label("song-title")
        details_box.add(self.song_title)

        self.album_name = make_label("album-name")
        details_box.add(self.album_name)

        self.artist_name = make_label("artist-name")
        details_box.add(self.artist_name)

        details_box.pack_start(Gtk.Box(), True, True, 0)
        box.pack_start(details_box, False, False, 5)

        return box

    def create_playback_controls(self) -> Gtk.Box:
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Scrubber and song progress/length labels
        scrubber_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.song_progress_label = Gtk.Label(label="-:--")
        scrubber_box.pack_start(self.song_progress_label, False, False, 5)

        self.song_scrubber = Gtk.Scale.new_with_range(
            orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=5
        )
        self.song_scrubber.set_name("song-scrubber")
        self.song_scrubber.set_draw_value(False)
        self.song_scrubber.set_restrict_to_fill_level(False)
        self.song_scrubber.connect(
            "change-value", lambda s, t, v: self.emit("song-scrub", v)
        )
        scrubber_box.pack_start(self.song_scrubber, True, True, 0)

        self.song_duration_label = Gtk.Label(label="-:--")
        scrubber_box.pack_start(self.song_duration_label, False, False, 5)

        box.add(scrubber_box)

        buttons = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        buttons.pack_start(Gtk.Box(), True, True, 0)

        # Repeat button
        repeat_button_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.repeat_button = IconToggleButton(
            "media-playlist-repeat", "Switch between repeat modes"
        )
        self.repeat_button.set_action_name("app.repeat-press")
        repeat_button_box.pack_start(Gtk.Box(), True, True, 0)
        repeat_button_box.pack_start(self.repeat_button, False, False, 0)
        repeat_button_box.pack_start(Gtk.Box(), True, True, 0)
        buttons.pack_start(repeat_button_box, False, False, 5)

        # Previous button
        self.prev_button = IconButton(
            "media-skip-backward-symbolic",
            "Go to previous song",
            icon_size=Gtk.IconSize.LARGE_TOOLBAR,
        )
        self.prev_button.set_action_name("app.prev-track")
        buttons.pack_start(self.prev_button, False, False, 5)

        # Play button
        self.play_button = IconButton(
            "media-playback-start-symbolic",
            "Play",
            relief=True,
            icon_size=Gtk.IconSize.LARGE_TOOLBAR,
        )
        self.play_button.set_name("play-button")
        self.play_button.set_action_name("app.play-pause")
        buttons.pack_start(self.play_button, False, False, 0)

        # Next button
        self.next_button = IconButton(
            "media-skip-forward-symbolic",
            "Go to next song",
            icon_size=Gtk.IconSize.LARGE_TOOLBAR,
        )
        self.next_button.set_action_name("app.next-track")
        buttons.pack_start(self.next_button, False, False, 5)

        # Shuffle button
        shuffle_button_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.shuffle_button = IconToggleButton(
            "media-playlist-shuffle-symbolic", "Toggle playlist shuffling"
        )
        self.shuffle_button.set_action_name("app.shuffle-press")
        shuffle_button_box.pack_start(Gtk.Box(), True, True, 0)
        shuffle_button_box.pack_start(self.shuffle_button, False, False, 0)
        shuffle_button_box.pack_start(Gtk.Box(), True, True, 0)
        buttons.pack_start(shuffle_button_box, False, False, 5)

        buttons.pack_start(Gtk.Box(), True, True, 0)
        box.add(buttons)

        return box

    def create_play_queue_volume(self) -> Gtk.Box:
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(Gtk.Box(), True, True, 0)
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        # Device button (for chromecast)
        self.device_button = IconButton(
            "chromecast-symbolic",
            "Show available audio output devices",
            icon_size=Gtk.IconSize.LARGE_TOOLBAR,
        )
        self.device_button.connect("clicked", self.on_device_click)
        box.pack_start(self.device_button, False, True, 5)

        self.device_popover = Gtk.PopoverMenu(modal=False, name="device-popover")
        self.device_popover.set_relative_to(self.device_button)

        device_popover_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            name="device-popover-box",
        )
        device_popover_header = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.popover_label = Gtk.Label(
            label="<b>Devices</b>",
            use_markup=True,
            halign=Gtk.Align.START,
            margin=5,
        )
        device_popover_header.add(self.popover_label)

        refresh_devices = IconButton("view-refresh-symbolic", "Refresh device list")
        refresh_devices.set_action_name("app.refresh-devices")
        device_popover_header.pack_end(refresh_devices, False, False, 0)

        device_popover_box.add(device_popover_header)

        device_list_and_loading = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.device_list = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        device_list_and_loading.add(self.device_list)

        device_popover_box.pack_end(device_list_and_loading, True, True, 0)

        self.device_popover.add(device_popover_box)

        # Play Queue button
        self.play_queue_button = IconButton(
            "view-list-symbolic",
            "Open play queue",
            icon_size=Gtk.IconSize.LARGE_TOOLBAR,
        )
        self.play_queue_button.connect("clicked", self.on_play_queue_click)
        box.pack_start(self.play_queue_button, False, True, 5)

        self.play_queue_popover = Gtk.PopoverMenu(modal=False, name="up-next-popover")
        self.play_queue_popover.set_relative_to(self.play_queue_button)

        play_queue_popover_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        play_queue_popover_header = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.popover_label = Gtk.Label(
            label="<b>Play Queue</b>",
            use_markup=True,
            halign=Gtk.Align.START,
            margin=10,
        )
        play_queue_popover_header.add(self.popover_label)

        self.load_play_queue_button = IconButton(
            "folder-download-symbolic", "Load Queue from Server", margin=5
        )
        self.load_play_queue_button.set_action_name("app.update-play-queue-from-server")
        play_queue_popover_header.pack_end(self.load_play_queue_button, False, False, 0)

        play_queue_popover_box.add(play_queue_popover_header)

        play_queue_loading_overlay = Gtk.Overlay()
        play_queue_scrollbox = Gtk.ScrolledWindow(
            min_content_height=600,
            min_content_width=400,
        )

        self.play_queue_store = Gtk.ListStore(
            bool,  # playable
            str,  # image filename
            str,  # title, album, artist
            bool,  # playing
            str,  # song ID
        )
        self.play_queue_list = Gtk.TreeView(
            model=self.play_queue_store,
            reorderable=True,
            headers_visible=False,
        )
        selection = self.play_queue_list.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        selection.set_select_function(lambda _, model, path, current: model[path[0]][0])

        # Album Art column. This function defines what image to use for the play queue
        # song icon.
        def filename_to_pixbuf(
            column: Any,
            cell: Gtk.CellRendererPixbuf,
            model: Gtk.ListStore,
            tree_iter: Gtk.TreeIter,
            flags: Any,
        ):
            cell.set_property("sensitive", model.get_value(tree_iter, 0))
            filename = model.get_value(tree_iter, 1)
            if not filename:
                cell.set_property("icon_name", "")
                return

            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(filename, 50, 50, True)

            # If this is the playing song, then overlay the play icon.
            if model.get_value(tree_iter, 3):
                play_overlay_pixbuf = GdkPixbuf.Pixbuf.new_from_file(
                    str(resolve_path("ui/images/play-queue-play.png"))
                )

                play_overlay_pixbuf.composite(
                    pixbuf, 0, 0, 50, 50, 0, 0, 1, 1, GdkPixbuf.InterpType.NEAREST, 200
                )

            cell.set_property("pixbuf", pixbuf)

        renderer = Gtk.CellRendererPixbuf()
        renderer.set_fixed_size(55, 60)
        column = Gtk.TreeViewColumn("", renderer)
        column.set_cell_data_func(renderer, filename_to_pixbuf)
        column.set_resizable(True)
        self.play_queue_list.append_column(column)

        renderer = Gtk.CellRendererText(markup=True, ellipsize=Pango.EllipsizeMode.END)
        column = Gtk.TreeViewColumn("", renderer, markup=2, sensitive=0)
        self.play_queue_list.append_column(column)

        self.play_queue_list.connect("row-activated", self.on_song_activated)
        self.play_queue_list.connect(
            "button-press-event", self.on_play_queue_button_press
        )

        # Set up drag-and-drop on the song list for editing the order of the
        # playlist.
        self.play_queue_store.connect("row-inserted", self.on_play_queue_model_row_move)
        self.play_queue_store.connect("row-deleted", self.on_play_queue_model_row_move)

        play_queue_scrollbox.add(self.play_queue_list)
        play_queue_loading_overlay.add(play_queue_scrollbox)

        self.play_queue_spinner = Gtk.Spinner(
            name="play-queue-spinner",
            active=False,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        play_queue_loading_overlay.add_overlay(self.play_queue_spinner)
        play_queue_popover_box.pack_end(play_queue_loading_overlay, True, True, 0)

        self.play_queue_popover.add(play_queue_popover_box)

        # Volume mute toggle
        self.volume_mute_toggle = IconButton(
            "audio-volume-high-symbolic", "Toggle mute"
        )
        self.volume_mute_toggle.set_action_name("app.mute-toggle")
        box.pack_start(self.volume_mute_toggle, False, True, 0)

        # Volume slider
        self.volume_slider = Gtk.Scale.new_with_range(
            orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=5
        )
        self.volume_slider.set_name("volume-slider")
        self.volume_slider.set_draw_value(False)
        self.volume_slider.connect("value-changed", self.on_volume_change)
        box.pack_start(self.volume_slider, True, True, 0)

        vbox.pack_start(box, False, True, 0)
        vbox.pack_start(Gtk.Box(), True, True, 0)
        return vbox
